import sys

import jax.numpy as jnp
import numpy

from .. import tensor as T
from .base import jax_wrap
import jax


# Add the apodization windows

names = ["blackman", "bartlett", "hamming", "hanning", "kaiser"]

module = sys.modules[__name__]
for name in names:
    module.__dict__.update({name: jax_wrap(jnp.__dict__[name])})


for name in ["convolve", "convolve2d", "correlate", "correlate2d"]:
    module.__dict__.update({name: jax_wrap(jax.scipy.signal.__dict__[name])})


# Add some utility functions


def fourier_complex_morlet(bandwidths, centers, N):
    """Complex Morlet wavelet in Fourier

    Parameters
    ----------

    bandwidths: array
        the bandwidth of the wavelet

    centers: array
        the centers of the wavelet

    freqs: array (optional)
        the frequency sampling in radion going from 0 to pi and back to 0
        :param N:

    """

    freqs = T.linspace(0, 2 * numpy.pi, N)
    envelop = T.exp(-0.25 * (freqs - centers) ** 2 * bandwidths ** 2)
    H = (freqs <= numpy.pi).astype("float32")
    return envelop * H


def complex_morlet(bandwidths, centers, time=None):
    """Complex Morlet wavelet

    It corresponds to with (B, C)::

        \phi(t) = \frac{1}{\pi B} e^{-\frac{t^2}{B}}e^{j2\pi C t}

    For a filter bank do

    J = 8
    Q = 1
    scales = T.power(2,T.linspace(0, J, J*Q))
    scales = scales[:, None]
    complex_morlet(scales, 1/scales)

    Parameters
    ----------

    bandwidths: array
        the bandwidth of the wavelet

    centers: array
        the centers of the wavelet

    time: array (optional)
        the time sampling

    Returns
    -------

    wavelet: array like
        the wavelet centered at 0

    """
    if time is None:
        B = 6 * bandwidths.max() + 1
        time = T.linspace(-(B // 2), B // 2, int(T.get(B)))
    envelop = T.exp(-((time / bandwidths) ** 2))
    wave = T.exp(1j * centers * time)
    return envelop * wave


def littewood_paley_normalization(filter_bank, down=None, up=None):
    lp = T.abs(filter_bank).sum(0)
    freq = T.linspace(0, 2 * numpy.pi, lp.shape[0])
    down = 0 if down is None else down
    up = numpy.pi or up
    lp = T.where(T.logical_and(freq >= down, freq <= up), lp, 1)
    return filter_bank / lp


def tukey(M, alpha=0.5):
    r"""Return a Tukey window, also known as a tapered cosine window.
    Parameters
    ----------
    M : int
        Number of points in the output window. If zero or less, an empty
        array is returned.
    alpha : float, optional
        Shape parameter of the Tukey window, representing the fraction of the
        window inside the cosine tapered region.
        If zero, the Tukey window is equivalent to a rectangular window.
        If one, the Tukey window is equivalent to a Hann window.
    Returns
    -------
    w : ndarray
        The window, with the maximum value normalized to 1 (though the value 1
        does not appear if `M` is even and `sym` is True).
    References
    ----------
    .. [1] Harris, Fredric J. (Jan 1978). "On the use of Windows for Harmonic
           Analysis with the Discrete Fourier Transform". Proceedings of the
           IEEE 66 (1): 51-83. :doi:`10.1109/PROC.1978.10837`
    .. [2] Wikipedia, "Window function",
           https://en.wikipedia.org/wiki/Window_function#Tukey_window
    """
    n = T.arange(0, M)
    width = int(numpy.floor(alpha * (M - 1) / 2.0))
    n1 = n[0 : width + 1]
    n2 = n[width + 1 : M - width - 1]
    n3 = n[M - width - 1 :]

    w1 = 0.5 * (1 + T.cos(numpy.pi * (-1 + 2.0 * n1 / alpha / (M - 1))))
    w2 = T.ones(n2.shape)
    w3 = 0.5 * (1 + T.cos(numpy.pi * (-2.0 / alpha + 1 + 2.0 * n3 / alpha / (M - 1))))

    w = T.concatenate((w1, w2, w3))

    return w


def morlet(M, s, w=5):
    """
    Complex Morlet wavelet.
    Parameters
    ----------
    M : int
        Length of the wavelet.

    s : float, optional
        Scaling factor, windowed from ``-s*2*pi`` to ``+s*2*pi``. Default is 1.
    w : float, optional
        Omega0. Default is 5
    complete : bool, optional
        Whether to use the complete or the standard version.
    Returns
    -------
    morlet : (M,) ndarray
    See Also
    --------
    morlet2 : Implementation of Morlet wavelet, compatible with `cwt`.
    scipy.signal.gausspulse
    Notes
    -----
    The standard version::
        pi**-0.25 * exp(1j*w*x) * exp(-0.5*(x**2))
    This commonly used wavelet is often referred to simply as the
    Morlet wavelet.  Note that this simplified version can cause
    admissibility problems at low values of `w`.
    The complete version::
        pi**-0.25 * (exp(1j*w*x) - exp(-0.5*(w**2))) * exp(-0.5*(x**2))
    This version has a correction
    term to improve admissibility. For `w` greater than 5, the
    correction term is negligible.
    Note that the energy of the return wavelet is not normalised
    according to `s`.
    The fundamental frequency of this wavelet in Hz is given
    by ``f = 2*s*w*r / M`` where `r` is the sampling rate.
    """
    limit = 2 * numpy.pi
    x = T.linspace(-limit, limit, M) * s
    sine = T.cos(w * x) + 1j * T.sin(w * x)
    envelop = T.exp(-0.5 * (x ** 2))

    # apply correction term for admissibility
    wave = sine - T.exp(-0.5 * (w ** 2))

    # now localize the wave to obtain a wavelet
    wavelet = wave * envelop * numpy.pi ** (-0.25)

    return wavelet


def bin_to_freq(bins, max_f):
    return (bins / bins.max()) * max_f


def freq_to_bin(freq, n_bins, fmin, fmax):
    unit = (fmax - fmin) / n_bins
    return (freq / unit).astype("int32")


def mel_to_freq(m, option="linear"):
    # convert mel to frequency with
    if option == "linear":
        # Fill in the linear scale
        f_sp = 200.0 / 3

        # And now the nonlinear scale
        # beginning of log region (Hz)
        min_log_hz = 1000.0
        min_log_mel = min_log_hz / f_sp  # same (Mels)
        # step size for log region
        logstep = numpy.log(6.4) / 27.0

        # If we have vector data, vectorize
        freq = min_log_hz * T.exp(logstep * (m - min_log_mel))
        return T.where(m >= min_log_mel, freq, f_sp * m)
    else:
        return 700 * (T.power(10.0, (m / 2595)) - 1)


def freq_to_mel(f, option="linear"):
    # convert frequency to mel with
    if option == "linear":

        # linear part slope
        f_sp = 200.0 / 3

        # Fill in the log-scale part
        min_log_hz = 1000.0  # beginning of log region (Hz)
        min_log_mel = min_log_hz / f_sp  # same (Mels)
        logstep = numpy.log(6.4) / 27.0  # step size for log region
        mel = min_log_mel + T.log(f / min_log_hz) / logstep
        return T.where(f >= min_log_hz, mel, f / f_sp)
    else:
        return 2595 * T.log10(1 + f / 700)


def power_to_db(S, ref=1.0, amin=1e-10, top_db=80.0):
    """Convert a power spectrogram (amplitude squared) to decibel (dB) units.

    https://librosa.github.io/librosa/_modules/librosa/core/spectrum.html#power_to_db.
    This computes the scaling ``10 * log10(S / ref)`` in a numerically
    stable way.

    Parameters
    ----------
    S : numpy.ndarray
        inumpy.t power

    ref : scalar or callable
        If scalar, the amplitude `abs(S)` is scaled relative to `ref`:
        `10 * log10(S / ref)`.
        Zeros in the output correspond to positions where `S == ref`.

        If callable, the reference value is computed as `ref(S)`.

    amin : float > 0 [scalar]
        minimum threshold for `abs(S)` and `ref`

    top_db : float >= 0 [scalar]
        threshold the output at `top_db` below the peak:
        ``max(10 * log10(S)) - top_db``

    Returns
    -------
    S_db : numpy.ndarray
        ``S_db ~= 10 * log10(S) - 10 * log10(ref)``

    See Also
    --------
    perceptual_weighting
    db_to_power
    amplitude_to_db
    db_to_amplitude
    """
    ref_value = numpy.abs(ref)
    log_spec = 10.0 * T.log10(T.maximum(amin, S) / T.maximum(amin, ref))
    if top_db is not None:
        if top_db < 0:
            error
        return T.maximum(log_spec, log_spec.max() - top_db)
    else:
        return log_spec


# Now some filter-bank and additional Time-Frequency Repr.


def sinc_bandpass(time, f0, f1):
    """
    ensure that f0<f1 and f0>0, f1<1
    whenever time is ..., -1, 0, 1, ...
    """
    high = f0 * T.sinc(time * f0)
    low = f1 * T.sinc(time * f1)
    return high - low


def mel_filterbank(length, n_filter, low, high, nyquist):
    # convert the low and high frequency into mel scale
    low_freq_mel = freq_to_mel(low)
    high_freq_mel = freq_to_mel(high)

    # generate center frequencies uniformly spaced in Mel scale.
    mel_points = T.linspace(low_freq_mel, high_freq_mel, n_filter + 2)
    # turn them into frequency and thus becoming in log scale
    freq_points = freq_to_bin(mel_to_freq(mel_points), length, 0, nyquist)
    peaks = T.expand_dims(freq_points, 1)
    freqs = T.range(length)
    filter_bank = T.hat_1D(freqs, peaks[:-2], peaks[1:-1], peaks[2:])
    return filter_bank


def stft(signal, window, hop, apod=T.ones, nfft=None, mode="valid"):
    """
    Compute the Shoft-Time-Fourier-Transform of a
    signal given the window length, hop and additional
    parameters.

    Parameters
    ----------

        signal: array
            the signal (possibly stacked of signals)

        window: int
            the window length to be considered for the fft

        hop: int
            the amount by which the window is moved

        apod: func
            a function that takes an integer as inumpy.t and return
            the apodization window of the same length

        nfft: int (optional)
            the number of bin that the fft on the window will use.
            If not given it is set the same as window.

        mode: 'valid', 'same' or 'full'
            the padding of the inumpy.t signals

    Returns
    -------

        output: complex array
            the complex stft
    """
    assert signal.ndim == 3
    if nfft is None:
        nfft = window
    if mode == "same":
        left = (window + 1) // 2
        psignal = T.pad(signal, [[0, 0], [0, 0], [left, window + 1 - left]])
    elif mode == "full":
        left = (window + 1) // 2
        psignal = T.pad(signal, [[0, 0], [0, 0], [window - 1, window - 1]])
    else:
        psignal = signal

    apodization = apod(window).reshape((1, 1, -1))

    p = T.extract_signal_patches(psignal, window, hop) * apodization
    assert nfft >= window
    pp = T.pad(p, [[0, 0], [0, 0], [0, 0], [0, nfft - window]])
    S = fft(pp)
    return S[..., : int(numpy.ceil(nfft / 2))].transpose([0, 1, 3, 2])


def spectrogram(signal, window, hop, apod=hanning, nfft=None, mode="valid"):
    return T.abs(stft(signal, window, hop, apod, nfft, mode))


def melspectrogram(
    signal,
    window,
    hop,
    n_filter,
    low_freq,
    high_freq,
    nyquist,
    nfft=None,
    mode="valid",
    apod=hanning,
):
    spec = spectrogram(signal, window, hop, apod, nfft, mode)
    filterbank = mel_filterbank(spec.shape[-2], n_filter, low_freq, high_freq, nyquist)
    flip_filterbank = filterbank.expand_dims(-1)
    output = (T.expand_dims(spec, -3) * flip_filterbank).sum(-2)
    return output


def mfcc(
    signal,
    window,
    hop,
    n_filter,
    low_freq,
    high_freq,
    nyquist,
    n_mfcc,
    nfft=None,
    mode="valid",
    apod=hanning,
):
    """
    https://librosa.github.io/librosa/_modules/librosa/feature/spectral.html#mfcc
    """
    tf = melspectrogram(
        signal, window, hop, n_filter, low_freq, high_freq, nyquist, nfft, mode, apod
    )
    tf_db = power_to_db(tf)
    M = dct(tf_db, axes=(2,))
    return M


def wvd(signal, window, hop, L, apod=hanning, mode="valid"):
    # define the following constant for clarity
    PI = 2 * 3.14159

    # compute the stft with 2 times bigger window to interp.
    s = stft(signal, window, hop, apod, nfft=2 * window, mode=mode)
    print("s", s.shape)
    # remodulate the stft prior the spectral correlation for simplicity
    # with the following mask
    step = 1 / window
    freq = T.linspace(-step * L, step * L, 2 * L + 1)
    time = T.range(s.shape[-1]).reshape((-1, 1))
    mask = T.complex(T.cos(PI * time * freq), T.sin(PI * time * freq)) * hanning(
        2 * L + 1
    )

    # extract vertical (freq) partches to perform auto correlation
    patches = T.extract_image_patches(s, (2 * L + 1, 1), (2, 1), mode="same")[
        ..., 0
    ]  # (N C F' T L)
    output = (patches * T.conj(T.flip(patches, -1)) * mask).sum(-1)
    return T.real(output)


def dct(signal, axes=(-1,)):
    """
    https://dsp.stackexchange.com/questions/2807/fast-cosine-transform-via-fft
    """
    if len(axes) > 1:
        raise NotImplemented("not yet implemented more than 1D")
    to_pad = [
        (0, 0) if ax not in axes else (0, signal.shape[ax]) for ax in range(signal.ndim)
    ]
    pad_signal = T.pad(signal, to_pad)
    exp = 2 * T.exp(-1j * 3.14159 * T.linspace(0, 0.5, signal.shape[axes[0]]))
    y = fft(pad_signal, axes=axes)
    cropped_y = T.dynamic_slice_in_dim(y, 0, signal.shape[axes[0]], axes[0])
    return T.real(cropped_y * exp.expand_dims(-1))


def phase_vocoder(D, rate, hop_length=None):
    """Phase vocoder.  Given an STFT matrix D, speed up by a factor of `rate`

    Based on the implementation provided by [1]_.

    .. note:: This is a simplified implementation, intended primarily for
             reference and pedagogical purposes.  It makes no attempt to
             handle transients, and is likely to produce many audible
             artifacts.  For a higher quality implementation, we recommend
             the RubberBand library [2]_ and its Python wrapper `pyrubberband`.

    .. [1] Ellis, D. P. W. "A phase vocoder in Matlab."
        Columbia University, 2002.
        http://www.ee.columbia.edu/~dpwe/resources/matlab/pvoc/

    .. [2] https://breakfastquay.com/rubberband/

    Examples
   --------
    >>> # Play at double speed
    >>> y, sr   = librosa.load(librosa.util.example_audio_file())
    >>> D       = librosa.stft(y, n_fft=2048, hop_length=512)
    >>> D_fast  = librosa.phase_vocoder(D, 2.0, hop_length=512)
    >>> y_fast  = librosa.istft(D_fast, hop_length=512)

    >>> # Or play at 1/3 speed
    >>> y, sr   = librosa.load(librosa.util.example_audio_file())
    >>> D       = librosa.stft(y, n_fft=2048, hop_length=512)
    >>> D_slow  = librosa.phase_vocoder(D, 1./3, hop_length=512)
    >>> y_slow  = librosa.istft(D_slow, hop_length=512)

    Parameters
    ----------
    D : numpy.ndarray [shape=(d, t), dtype=complex]
        STFT matrix

    rate :  float > 0 [scalar]
        Speed-up factor: `rate > 1` is faster, `rate < 1` is slower.

    hop_length : int > 0 [scalar] or None
        The number of samples between successive columns of `D`.

        If None, defaults to `n_fft/4 = (D.shape[0]-1)/2`

    Returns
    -------
    D_stretched : numpy.ndarray [shape=(d, t / rate), dtype=complex]
        time-stretched STFT

    """
    n_fft = 2 * (D.shape[0] - 1)
    if hop_length is None:
        hop_length = int(n_fft // 4)

    time_steps = numpy.arange(0, D.shape[1], rate, "float32")

    # Create an empty output array
    d_stretch = T.zeros((D.shape[0], len(time_steps)), D.dtype)

    # Expected phase advance in each bin
    phi_advance = T.linspace(0, numpy.pi * hop_length, D.shape[0])

    # Pad 0 columns to simplify boundary logic
    D = T.pad(D, [(0, 0), (0, 2)], mode="constant")
    D = D[:, time_steps.astype("int32")]

    alpha = numpy.mod(time_steps, 1.0)

    mag = (1.0 - alpha) * numpy.abs(D[:, :-1]) + alpha * numpy.abs(D[:, 1:])

    # Compute phase advance
    dphase = numpy.angle(D[:, 1:]) - numpy.angle(D[:, :-1]) - phi_advance[:, None]
    # Wrap to -pi:pi range
    dphase = dphase - 2.0 * numpy.pi * numpy.round(dphase / (2.0 * numpy.pi))

    # Phase accumulator; initialize to the first sample
    phase_acc = T.concatenate(
        [numpy.angle(D[:, [0]]), T.cumsum(phi_advance + dphase, 1)], 1
    )

    d_stretch = mag * T.complex(T.cos(phase_acc), T.sin(phase_acc))

    return d_stretch


def istft(
    stft_matrix,
    hop_length=None,
    win_length=None,
    window="hann",
    center=True,
    dtype=numpy.float32,
    length=None,
):
    """
    Inverse short-time Fourier transform (ISTFT).

    Converts a complex-valued spectrogram `stft_matrix` to time-series `y`
    by minimizing the mean squared error between `stft_matrix` and STFT of
    `y` as described in [1]_ up to Section 2 (reconstruction from MSTFT).

    In general, window function, hop length and other parameters should be same
    as in stft, which mostly leads to perfect reconstruction of a signal from
    unmodified `stft_matrix`.

    .. [1] D. W. Griffin and J. S. Lim,
        "Signal estimation from modified short-time Fourier transform,"
        IEEE Trans. ASSP, vol.32, no.2, pp.236–243, Apr. 1984.

    Parameters
    ----------
    stft_matrix : numpy.ndarray [shape=(1 + n_fft/2, t)]
        STFT matrix from `stft`

    hop_length : int > 0 [scalar]
        Number of frames between STFT columns.
        If unspecified, defaults to `win_length / 4`.

    win_length : int <= n_fft = 2 * (stft_matrix.shape[0] - 1)
        When reconstructing the time series, each frame is windowed
        and each sample is normalized by the sum of squared window
        according to the `window` function (see below).

        If unspecified, defaults to `n_fft`.

    window : string, tuple, number, function, numpy.ndarray [shape=(n_fft,)]
        - a window specification (string, tuple, or number);
          see `scipy.signal.get_window`
        - a window function, such as `scipy.signal.hanning`
        - a user-specified window vector of length `n_fft`

        .. see also:: `filters.get_window`

    center : boolean
        - If `True`, `D` is assumed to have centered frames.
        - If `False`, `D` is assumed to have left-aligned frames.

    dtype : numeric type
        Real numeric type for `y`.  Default is 32-bit float.

    length : int > 0, optional
        If provided, the output `y` is zero-padded or clipped to exactly
        `length` samples.

    Returns
    -------
    y : numpy.ndarray [shape=(n,)]
        time domain signal reconstructed from `stft_matrix`

    See Also
    --------
    stft : Short-time Fourier Transform

    Notes
    -----
    This function caches at level 30.

    Examples
    --------
    >>> y, sr = librosa.load(librosa.util.example_audio_file())
    >>> D = librosa.stft(y)
    >>> y_hat = librosa.istft(D)
    >>> y_hat
    array([ -4.812e-06,  -4.267e-06, ...,   6.271e-06,   2.827e-07], dtype=float32)

    Exactly preserving length of the inumpy.t signal requires explicit padding.
    Otherwise, a partial frame at the end of `y` will not be represented.

    >>> n = len(y)
    >>> n_fft = 2048
    >>> y_pad = librosa.util.fix_length(y, n + n_fft // 2)
    >>> D = librosa.stft(y_pad, n_fft=n_fft)
    >>> y_out = librosa.istft(D, length=n)
    >>> numpy.max(numpy.abs(y - y_out))
    1.4901161e-07
    """

    n_fft = 2 * (stft_matrix.shape[0] - 1)

    # By default, use the entire frame
    if win_length is None:
        win_length = n_fft

    # Set the default hop, if it's not already specified
    if hop_length is None:
        hop_length = int(win_length // 4)

    ifft_window = get_window(window, win_length, fftbins=True)

    # Pad out to match n_fft, and add a broadcasting axis
    ifft_window = util.pad_center(ifft_window, n_fft)[:, numpy.newaxis]

    # For efficiency, trim STFT frames according to signal length if available
    if length:
        if center:
            padded_length = length + int(n_fft)
        else:
            padded_length = length
        n_frames = min(
            stft_matrix.shape[1], int(numpy.ceil(padded_length / hop_length))
        )
    else:
        n_frames = stft_matrix.shape[1]

    expected_signal_len = n_fft + hop_length * (n_frames - 1)
    y = numpy.zeros(expected_signal_len, dtype=dtype)

    n_columns = int(util.MAX_MEM_BLOCK // (stft_matrix.shape[0] * stft_matrix.itemsize))

    fft = get_fftlib()

    frame = 0
    for bl_s in range(0, n_frames, n_columns):
        bl_t = min(bl_s + n_columns, n_frames)

        # invert the block and apply the window function
        ytmp = ifft_window * fft.irfft(stft_matrix[:, bl_s:bl_t], axis=0)

        # Overlap-add the istft block starting at the i'th frame
        __overlap_add(y[frame * hop_length :], ytmp, hop_length)

        frame += bl_t - bl_s

    # Normalize by sum of squared window
    ifft_window_sum = window_sumsquare(
        window,
        n_frames,
        win_length=win_length,
        n_fft=n_fft,
        hop_length=hop_length,
        dtype=dtype,
    )

    approx_nonzero_indices = ifft_window_sum > util.tiny(ifft_window_sum)
    y[approx_nonzero_indices] /= ifft_window_sum[approx_nonzero_indices]

    if length is None:
        # If we don't need to control length, just do the usual center trimming
        # to eliminate padded data
        if center:
            y = y[int(n_fft // 2) : -int(n_fft // 2)]
    else:
        if center:
            # If we're centering, crop off the first n_fft//2 samples
            # and then trim/pad to the target length.
            # We don't trim the end here, so that if the signal is zero-padded
            # to a longer duration, the decay is smooth by windowing
            start = int(n_fft // 2)
        else:
            # If we're not centering, start at 0 and trim/pad as necessary
            start = 0

        y = util.fix_length(y[start:], length)

    return y


def hilbert_transform(signal):
    """
    the time should be the last dimension
    return the analytical signal
    """
    M = signal.shape[-1]
    heavyside = T.array([1, 0], dtype="float32").repeat(M // 2)
    mask = T.index_add(T.ones(M), T.index[..., 1 : M // 2], 1)
    return T.signal.ifft(T.signal.fft(signal) * mask * heavyside)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = "Randall Balestriero"

import io
import os
import pickle, gzip
import urllib.request
import numpy as np
import time
import zipfile
from tqdm import tqdm
from scipy.io.wavfile import read as wav_read


class birdvox_dcase_20k:
    """Binary bird detection classification

    Dataset is 16.5Go compressed.

    BirdVox-DCASE-20k: a dataset for bird audio detection in 10-second
    clips
    
    Version 2.0, March 2018.

    Created By

    Vincent Lostanlen (1, 2, 3), Justin Salamon (2, 3), Andrew Farnsworth
    (1), Steve Kelling (1), and Juan Pablo Bello (2, 3).

    (1): Cornell Lab of Ornithology (CLO)
    (2): Center for Urban Science and Progress, New York University
    (3): Music and Audio Research Lab, New York University

    `link <https://wp.nyu.edu/birdvox>`_
     
    Description

    The BirdVox-DCASE-20k dataset contains 20,000 ten-second audio
    recordings. These recordings come from ROBIN autonomous recording
    units, placed near Ithaca, NY, USA during the fall 2015. They were
    captured on the night of September 23rd, 2015, by six different
    sensors, originally numbered 1, 2, 3, 5, 7, and 10.

    Out of these 20,000 recording, 10,017 (50.09%) contain at least one
    bird vocalization (either song, call, or chatter).

    The dataset is a derivative work of the BirdVox-full-night dataset
    [1], containing almost as much data but formatted into ten-second
    excerpts rather than ten-hour full night recordings.

    In addition, the BirdVox-DCASE-20k dataset is provided as a
    development set in the context of the "Bird Audio Detection"
    challenge, organized by DCASE (Detection and Classification of
    Acoustic Scenes and Events) and the IEEE Signal Processing Society.

    The dataset can be used, among other things, for the development and
    evaluation of bioacoustic classification models.


    We refer the reader to [1] for details on the distribution of the
    data and [2] for details on the hardware of ROBIN recording units.

    [1] V. Lostanlen, J. Salamon, A. Farnsworth, S. Kelling, J.P. Bello.
    "BirdVox-full-night: a dataset and benchmark for avian flight call
    detection", Proc. IEEE ICASSP, 2018.

    [2] J. Salamon, J. P. Bello, A. Farnsworth, M. Robbins, S. Keen,
    H. Klinck, and S. Kelling. Towards the Automatic Classification of
    Avian Flight Calls for Bioacoustic Monitoring. PLoS One, 2016.

    Data Files

    The wav folder contains the recordings as WAV files, sampled at
    44,1 kHz, with a single channel (mono). The original sample rate
    was 24 kHz.

    The name of each wav file is a random 128-bit UUID (Universal
    Unique IDentifier) string, which is randomized with respect to the
    origin of the recording in BirdVox-full-night, both in terms of
    time (UTC hour at the start of the excerpt) and space (location of
    the sensor).

    The origin of each 10-second excerpt is known by the challenge
    organizers, but not disclosed to the participants.

    Please Acknowledge BirdVox-DCASE-20k in Academic Research

    When BirdVox-70k is used for academic research, we would highly
    appreciate it if  scientific publications of works partly based on
    this dataset cite the following publication:

    V. Lostanlen, J. Salamon, A. Farnsworth, S. Kelling, J. Bello.
    "BirdVox-full-night: a dataset and benchmark for avian flight call
    detection", Proc. IEEE ICASSP, 2018.

    @inproceedings{lostanlen2018icassp,
    title = {BirdVox-full-night: a dataset and benchmark for avian
    flight call detection},
    author = {Lostanlen, Vincent and Salamon, Justin and Farnsworth,
    Andrew and Kelling, Steve and Bello, Juan Pablo},
    booktitle = {Proc. IEEE ICASSP},
    year = {2018},
    published = {IEEE},
    venue = {Calgary, Canada},
    month = {April},
    }

    The creation of this dataset was supported by NSF grants 1125098
    (BIRDCAST) and 1633259 (BIRDVOX), a Google Faculty Award, the Leon
    Levy Foundation, and two anonymous donors.
    """

    @staticmethod
    def download(path):
        """
        Download the Birdvox dataset and store the result into the given
        path

        Parameters
        ----------

            path: str
                the path where the downloaded files will be stored. If the
                directory does not exist, it is created.
        """

        # Check if directory exists
        if not os.path.isdir(path + "birdvox_dcase_20k"):
            print("Creating birdvox_dcase_20k Directory")
            os.mkdir(path + "birdvox_dcase_20k")
        base = "https://zenodo.org/record/1208080/files/"
        filename = "BirdVox-DCASE-20k.zip"
        if not os.path.exists(path + "birdvox_dcase_20k/" + filename):
            url = base + filename + "?download=1"
            urllib.request.urlretrieve(url, path + "birdvox_dcase_20k/" + filename)
        url = "https://ndownloader.figshare.com/files/10853300"
        filename = "data_labels.csv"
        if not os.path.exists(path + "birdvox_dcase_20k/" + filename):
            urllib.request.urlretrieve(url, path + "birdvox_dcase_20k/" + filename)

    @staticmethod
    def load(path=None):
        """
        Parameters
        ----------
            path: str (optional)
                default ($DATASET_PATH), the path to look for the data and
                where the data will be downloaded if not present

        Returns
        -------

            wavs: array
                the waveforms in the time amplitude domain

            labels: array
                binary values representing the presence or not of an avian

            recording: array
                the file number from which the sample has been extracted

        """

        if path is None:
            path = os.environ["DATASET_PATH"]

        birdvox_dcase_20k.download(path)

        t0 = time.time()

        # Loading the file
        basefile = path + "birdvox_dcase_20k/BirdVox-DCASE-20k.zip"
        wavs = list()
        labels = np.loadtxt(
            path + "birdvox_dcase_20k/data_labels.csv",
            skiprows=1,
            delimiter=",",
            dtype="str",
        )
        wav_names = list(labels[:, 0])
        wav_labels = labels[:, 2].astype("int")
        labels = list()
        f = zipfile.ZipFile(basefile)
        for name in tqdm(f.namelist(), ascii=True):
            filename = name.split("/")[-1][:-4]
            if ".wav" not in name or filename not in wav_names:
                continue
            byt = io.BytesIO(f.read(name))
            wavs.append(wav_read(byt)[1].astype("float32"))
            labels.append(wav_labels[wav_names.index(filename)])

        wavs = np.array(wavs).astype("float32")
        labels = np.array(labels).astype("int32")

        print("Dataset birdvox_dcase_20k loaded in {0:.2f}s.".format(time.time() - t0))

        return wavs, labels



import urllib.request
import time
import sys
import os, sys, tarfile, io
import numpy as np
import matplotlib.pyplot as plt


class stl10:
    """ Image classification with extra unlabeled images.

    The `STL-10 <https://cs.stanford.edu/~acoates/stl10/>`_ dataset is an image 
    recognition dataset for developing unsupervised feature learning, 
    deep learning, self-taught learning algorithms. It is inspired by the 
    CIFAR-10 dataset but with 
    some modifications. In particular, each class has fewer labeled 
    training examples than in CIFAR-10, but a very 
    large set of unlabeled examples is provided to learn image models prior 
    to supervised training. The primary challenge is to make use of the 
    unlabeled data (which comes from a similar but different distribution from 
    the labeled data) to build a useful prior. We also expect that the higher 
    resolution of this dataset (96x96) will make it a challenging benchmark 
    for developing more scalable unsupervised learning methods.

    """

    classes = [
        "airplane",
        "bird",
        "car",
        "cat",
        "deer",
        "dog",
        "horse",
        "monkey",
        "ship",
        "truck",
    ]

    @staticmethod
    def download(path):
        # Check if directory exists
        if not os.path.isdir(path + "stl10"):
            print("\tCreating stl10 Directory")
            os.mkdir(path + "stl10")

        # Check if data file exists
        if not os.path.exists(path + "stl10/stl10_binary.tar.gz"):
            print("\tDownloading stl10 Dataset...")
            url = "http://ai.stanford.edu/~acoates/stl10/stl10_binary.tar.gz"
            urllib.request.urlretrieve(url, path + "stl10/stl10_binary.tar.gz")

    @staticmethod
    def load(path=None):
        """
        load the data

        Parameters
        ----------

        path: str (optional)
            the path to look for the data and where it will be downloaded if
            not present

        Returns
        -------

        train_images: array
            the training images

        train_labels: array
            the training labels

        test_images: array
            the test images

        test_labels: array
            the test labels

        extra_images: array
            the unlabeled additional images
        """
        if path is None:
            path = os.environ["DATASET_path"]

        # Load the dataset (download if necessary) and set
        # the class attributes.

        print("Loading stl10")
        t = time.time()

        stl10.download(path)

        # Loading Dataset
        file_ = tarfile.open(path + "stl10/stl10_binary.tar.gz", "r:gz")
        # loading test label
        read_file = file_.extractfile("stl10_binary/test_y.bin").read()
        test_y = np.frombuffer(io.BytesIO(read_file).read(), dtype=np.uint8) - 1
        # loading train label
        read_file = file_.extractfile("stl10_binary/train_y.bin").read()
        train_y = np.frombuffer(io.BytesIO(read_file).read(), dtype=np.uint8) - 1
        # load test images
        read_file = file_.extractfile("stl10_binary/test_X.bin").read()
        test_X = (
            np.frombuffer(io.BytesIO(read_file).read(), dtype=np.uint8)
            .reshape((-1, 3, 96, 96))
            .transpose([0, 1, 3, 2])
        )
        # load train images
        read_file = file_.extractfile("stl10_binary/train_X.bin").read()
        train_X = (
            np.frombuffer(io.BytesIO(read_file).read(), dtype=np.uint8)
            .reshape((-1, 3, 96, 96))
            .transpose([0, 1, 3, 2])
        )
        # load unlabelled images
        read_file = file_.extractfile("stl10_binary/unlabeled_X.bin").read()
        unlabeled_X = (
            np.frombuffer(io.BytesIO(read_file).read(), dtype=np.uint8)
            .reshape((-1, 3, 96, 96))
            .transpose([0, 1, 3, 2])
        )

        print("Dataset stl10 loaded in", "{0:.2f}".format(time.time() - t), "s.")
        return train_X, train_y, test_X, test_y, unlabeled_X



import os
import gzip
import urllib.request
import numpy as np
import time


class fashionmnist:
    """Grayscale image classification
    
    `Zalando <https://jobs.zalando.com/tech/>`_ 's article image classification.
    `Fashion-MNIST <https://github.com/zalandoresearch/fashion-mnist>`_ is 
    a dataset of `Zalando <https://jobs.zalando.com/tech/>`_ 's article 
    images consisting of a training set of 60,000 examples and a test set 
    of 10,000 examples. Each example is a 28x28 grayscale image, associated 
    with a label from 10 classes. We intend Fashion-MNIST to serve as a direct
    drop-in replacement for the original MNIST dataset for benchmarking 
    machine learning algorithms. It shares the same image size and structure 
    of training and testing splits.
    """

    def download(path):
        """
        Download the fashion-MNIST dataset and store the result into the given
        path

        Parameters
        ----------

            path: str
                the path where the downloaded files will be stored. If the
                directory does not exist, it is created.
        """

        print("Loading fashionmnist")

        t0 = time.time()

        if not os.path.isdir(path + "fashionmnist"):
            print("\tCreating fashionmnist Directory")
            os.mkdir(path + "fashionmnist")
        base_url = "http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/"
        if not os.path.exists(path + "fashionmnist/train-images.gz"):
            print("\tDownloading train images")
            url = base_url + "train-images-idx3-ubyte.gz"
            urllib.request.urlretrieve(url, path + "fashionmnist/train-images.gz")

        if not os.path.exists(path + "fashionmnist/train-labels.gz"):
            print("\tDownloading train labels")
            url = base_url + "train-labels-idx1-ubyte.gz"
            urllib.request.urlretrieve(url, path + "fashionmnist/train-labels.gz")

        if not os.path.exists(path + "fashionmnist/test-images.gz"):
            print("\tDownloading test images")
            url = base_url + "t10k-images-idx3-ubyte.gz"
            urllib.request.urlretrieve(url, path + "fashionmnist/test-images.gz")

        if not os.path.exists(path + "fashionmnist/test-labels.gz"):
            print("\tDownloading test labels")
            url = base_url + "t10k-labels-idx1-ubyte.gz"
            urllib.request.urlretrieve(url, path + "fashionmnist/test-labels.gz")

    def load(path=None):
        """
        Parameters
        ----------
            path: str (optional)
                default ($DATASET_PATH), the path to look for the data and
                where the data will be downloaded if not present

        Returns
        -------

            train_images: array

            train_labels: array

            test_images: array

            test_labels: array

        """
        if path is None:
            path = os.environ["DATASET_PATH"]
        # Loading the file
        fashionmnist.download(path)
        t0 = time.time()
        print("\tLoading fashionmnist")
        with gzip.open(path + "fashionmnist/train-labels.gz", "rb") as lbpath:
            train_labels = np.frombuffer(lbpath.read(), dtype=np.uint8, offset=8)

        with gzip.open(path + "fashionmnist/train-images.gz", "rb") as lbpath:
            train_images = np.frombuffer(lbpath.read(), dtype=np.uint8, offset=16)
        train_images = train_images.reshape((-1, 1, 28, 28)).astype("float32")

        with gzip.open(path + "fashionmnist/test-labels.gz", "rb") as lbpath:
            test_labels = np.frombuffer(lbpath.read(), dtype=np.uint8, offset=8)

        with gzip.open(path + "fashionmnist/test-images.gz", "rb") as lbpath:
            test_images = np.frombuffer(lbpath.read(), dtype=np.uint8, offset=16)
        test_images = test_images.reshape((-1, 1, 28, 28)).astype("float32")

        print("Dataset fashionmnist loaded in", "{0:.2f}s.".format(time.time() - t0))
        return (
            np.copy(train_images),
            np.copy(train_labels),
            np.copy(test_images),
            np.copy(test_labels),
        )



import os
import pickle, gzip
import urllib.request
import numpy as np
import time
import tarfile
from tqdm import tqdm
import zipfile
from scipy.io.wavfile import read as wav_read
import io


class esc:
    """ESC-10/50: Environmental Sound Classification

    https://github.com/karolpiczak/ESC-50#download

    The ESC-50 dataset is a labeled collection of 2000 environmental audio
    recordings suitable for benchmarking methods of environmental sound
    classification.

    The dataset consists of 5-second-long recordings organized into 50
    semantical classes (with 40 examples per class) loosely arranged into 5
    major categories:
        Animals
        Natural soundscapes & water sounds
        Human, non-speech sounds
        Interior/domestic sounds
        Exterior/urban noises

    Clips in this dataset have been manually extracted from public field
    recordings gathered by the Freesound.org project. The dataset has been
    prearranged into 5 folds for comparable cross-validation, making sure
    that fragments from the same original source file are contained in a
    single fold.

    """

    fine_to_coarse = {
        "dog": 0,
        "rooster": 0,
        "pig": 0,
        "cow": 0,
        "frog": 0,
        "cat": 0,
        "hen": 0,
        "insects": 0,
        "sheep": 0,
        "crow": 0,
        "rain": 1,
        "sea_waves": 1,
        "crackling_fire": 1,
        "crickets": 1,
        "chirping_birds": 1,
        "water_drops": 1,
        "wind": 1,
        "pouring_water": 1,
        "toilet_flush": 1,
        "thunderstorm": 1,
        "crying_baby": 2,
        "sneezing": 2,
        "clapping": 2,
        "breathing": 2,
        "coughing": 2,
        "footsteps": 2,
        "laughing": 2,
        "brushing_teeth": 2,
        "snoring": 2,
        "drinking_sipping": 2,
        "door_wood_knock": 3,
        "mouse_click": 3,
        "keyboard_typing": 3,
        "door_wood_creaks": 3,
        "can_opening": 3,
        "washing_machine": 3,
        "vacuum_cleaner": 3,
        "clock_alarm": 3,
        "clock_tick": 3,
        "glass_breaking": 3,
        "helicopter": 4,
        "chainsaw": 4,
        "siren": 4,
        "car_horn": 4,
        "engine": 4,
        "train": 4,
        "church_bells": 4,
        "airplane": 4,
        "fireworks": 4,
        "hand_saw": 4,
    }

    def download(path):

        # Check if directory exists
        if not os.path.isdir(path + "esc50"):
            print("\tCreating esc50 Directory")
            os.mkdir(path + "esc50")
        url = "https://github.com/karoldvl/ESC-50/archive/master.zip"
        # Check if file exists
        if not os.path.exists(path + "esc50/master.zip"):
            print("\tDownloading esc50")
            urllib.request.urlretrieve(url, path + "esc50/master.zip")

    def load(path=None):
        """ESC 50.
    
        https://github.com/karolpiczak/ESC-50#download
    
    
        Parameters
        ----------
        
        path: str (optional)
                default $DATASET_path), the path to look for the data and
                where the data will be downloaded if not present

        Returns
        -------

        wavs: array
            the wavs as a numpy array (matrix) with first dimension the data
            and second dimension time
        
        fine_labels: array
            the labels of the final classes (50 different ones) as a integer
            vector
        
        coarse_labels: array
            the labels of the classes big cateogry (5 of them)
        
        folds: array
            the fold as an integer from 1 to 5 specifying how to split the data
            one should not split a fold into train and set as it would
            make the same recording (but different subparts) be present in train
            and test, biasing optimistically the results.
        
        esc10: array
            the boolean vector specifying if the corresponding datum (wav, label,
            ...) is in the ESC-10 dataset or not. That is, to load the ESC-10
            dataset simply load ESC-50 and use this boolean vector to extract
            only the ESC-10 data.
        """

        if path is None:
            path = os.environ["DATASET_PATH"]
        esc.download(path)
        t0 = time.time()

        f = zipfile.ZipFile(path + "esc50/master.zip")

        meta = np.loadtxt(
            io.BytesIO(f.read("ESC-50-master/meta/esc50.csv")),
            delimiter=",",
            skiprows=1,
            dtype="str",
        )
        filenames = list(meta[:, 0])
        folds = meta[:, 1].astype("int32")
        fine_labels = meta[:, 2].astype("int32")
        categories = meta[:, 3]
        esc10 = meta[:, 4] == "True"
        coarse_labels = np.array([esc.fine_to_coarse[c] for c in categories])
        coarse_labels = coarse_labels.astype("int32")

        wavs = list()
        order = list()
        N = 0
        for filename in tqdm(f.namelist(), ascii=True):
            if ".wav" not in filename:
                continue
            wavfile = f.read(filename)
            byt = io.BytesIO(wavfile)
            wavs.append(wav_read(byt)[1].astype("float32"))
            order.append(filenames.index(filename.split("/")[-1]))
            N = max(N, len(wavs[-1]))

        all_wavs = np.zeros((len(wavs), N))
        for i in range(len(wavs)):
            left = (N - len(wavs[i])) // 2
            all_wavs[order[i], left : left + len(wavs[i])] = wavs[i]
        return all_wavs, fine_labels, coarse_labels, folds, esc10



import os
import pickle, gzip
import urllib.request
import numpy as np
import time


def download(path):
    """
    Download the MNIST dataset and store the result into the given
    path

    Parameters
    ----------

        path: str
            the path where the downloaded files will be stored. If the
            directory does not exist, it is created.
    """

    # Check if directory exists
    if not os.path.isdir(path + "mnist"):
        print("Creating mnist Directory")
        os.mkdir(path + "mnist")

    # Check if file exists
    if not os.path.exists(path + "mnist/mnist.pkl.gz"):
        td = time.time()
        print("Creating mnist")
        url = "http://deeplearning.net/data/mnist/mnist.pkl.gz"
        urllib.request.urlretrieve(url, path + "mnist/mnist.pkl.gz")


def load(path=None):
    """
    Parameters
    ----------
        path: str (optional)
            default ($DATASET_PATH), the path to look for the data and
            where the data will be downloaded if not present

    Returns
    -------

        train_images: array

        train_labels: array

        valid_images: array

        valid_labels: array

        test_images: array

        test_labels: array

    """

    if path is None:
        path = os.environ["DATASET_PATH"]

    download(path)

    t0 = time.time()

    # Loading the file
    print("Loading mnist")
    f = gzip.open(path + "mnist/mnist.pkl.gz", "rb")
    train_set, valid_set, test_set = pickle.load(f, encoding="latin1")
    f.close()

    train_set = (
        train_set[0].reshape((-1, 1, 28, 28)).astype("float32"),
        train_set[1].astype("int32"),
    )
    test_set = (
        test_set[0].reshape((-1, 1, 28, 28)).astype("float32"),
        test_set[1].astype("int32"),
    )
    valid_set = (
        valid_set[0].reshape((-1, 1, 28, 28)).astype("float32"),
        valid_set[1].astype("int32"),
    )

    data = {
        "train_set/images": train_set[0],
        "train_set/labels": train_set[1],
        "test_set/images": test_set[0],
        "test_set/labels": test_set[1],
        "valid_set/images": valid_set[0],
        "valid_set/labels": valid_set[1],
    }

    print("Dataset mnist loaded in {0:.2f}s.".format(time.time() - t0))

    return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import io
import os
import time
import zipfile
import matplotlib.image as mpimg
import urllib
import numpy as np


__author__ = "Randall Balestriero"


class ibeans:
    """Plant images classification.

    This dataset is of leaf images taken in the field in different
    districts in Uganda by the Makerere AI lab in collaboration with the
    National Crops Resources Research Institute (NaCRRI), the national
    body in charge of research in agriculture in Uganda.

    The goal is to build a robust machine learning model that is able
    to distinguish between diseases in the Bean plants. Beans are an
    important cereal food crop for Africa grown by many small-holder
    farmers - they are a significant source of proteins for school-age
    going children in East Africa.

    The data is of leaf images representing 3 classes: the healthy class of
    images, and two disease classes including Angular Leaf Spot and Bean
    Rust diseases. The model should be able to distinguish between these 3
    classes with high accuracy. The end goal is to build a robust, model
    that can be deployed on a mobile device and used in the field by a
    farmer.

    The data includes leaf images taken in the field. The figure above
    depicts examples of the types of images per class. Images were taken
    from the field/garden a basic smartphone.

    The images were then annotated by experts from NaCRRI who determined
    for each image which disease was manifested. The experts were part of
    the data collection team and images were annotated directly during the
    data collection process in the field.

    Class   Examples
    Healthy class   428
    Angular Leaf Spot   432
    Bean Rust   436
    Total:  1,296

    Data Released   20-January-2020
    License     MIT
    Credits     Makerere AI Lab
    """

    classes = ["angular_leaf_spot", "bean_rust", "healthy"]

    @staticmethod
    def download(path):
        """
        Download the ibeans dataset and store the result into the given
        path

        Parameters
        ----------

            path: str
                the path where the downloaded files will be stored. If the
                directory does not exist, it is created.
        """

        # Check if directory exists
        if not os.path.isdir(path + "ibeans"):
            print("Creating mnist Directory")
            os.mkdir(path + "ibeans")
        # Check if file exists
        if not os.path.exists(path + "ibeans/train.zip"):
            url = "https://storage.googleapis.com/ibeans/train.zip"
            urllib.request.urlretrieve(url, path + "ibeans/train.zip")

        # Check if file exists
        if not os.path.exists(path + "ibeans/test.zip"):
            url = "https://storage.googleapis.com/ibeans/test.zip"
            urllib.request.urlretrieve(url, path + "ibeans/test.zip")

        # Check if file exists
        if not os.path.exists(path + "ibeans/validation.zip"):
            url = "https://storage.googleapis.com/ibeans/validation.zip"
            urllib.request.urlretrieve(url, path + "ibeans/validation.zip")

    @staticmethod
    def load(path=None):
        """
        Parameters
        ----------
            path: str (optional)
                default ($DATASET_PATH), the path to look for the data and
                where the data will be downloaded if not present

        Returns
        -------

            train_images: array

            train_labels: array

            valid_images: array

            valid_labels: array

            test_images: array

            test_labels: array

        """

        if path is None:
            path = os.environ["DATASET_PATH"]

        ibeans.download(path)

        t0 = time.time()

        # Loading the file
        train_images = list()
        train_labels = list()
        f = zipfile.ZipFile(path + "ibeans/train.zip")
        for filename in f.namelist():
            if ".jpg" not in filename:
                continue
            train_images.append(mpimg.imread(io.BytesIO(f.read(filename)), "jpg"))
            train_labels.append(ibeans.classes.index(filename.split("/")[1]))

        # Loading the file
        test_images = list()
        test_labels = list()
        f = zipfile.ZipFile(path + "ibeans/test.zip")
        for filename in f.namelist():
            if ".jpg" not in filename:
                continue
            test_images.append(mpimg.imread(io.BytesIO(f.read(filename)), "jpg"))
            test_labels.append(ibeans.classes.index(filename.split("/")[1]))

        # Loading the file
        valid_images = list()
        valid_labels = list()
        f = zipfile.ZipFile(path + "ibeans/validation.zip")
        for filename in f.namelist():
            if ".jpg" not in filename:
                continue
            valid_images.append(mpimg.imread(io.BytesIO(f.read(filename)), "jpg"))
            valid_labels.append(ibeans.classes.index(filename.split("/")[1]))

        train_images = np.array(train_images)
        test_images = np.array(test_images)
        valid_images = np.array(valid_images)

        train_labels = np.array(train_labels)
        test_labels = np.array(test_labels)
        valid_labels = np.array(valid_labels)

        print("Dataset ibeans loaded in {0:.2f}s.".format(time.time() - t0))

        return (
            train_images,
            train_labels,
            valid_images,
            valid_labels,
            test_images,
            test_labels,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Base of symjax."""

import fnmatch

import jax
import numpy
from jax import jacfwd, jacrev

import symjax
from symjax import tensor as t
from symjax.tensor import random
import networkx as nx
import re


def natural_key(string_):
    """See http://www.codinghorror.com/blog/archives/001018.html"""
    return [int(s) if s.isdigit() else s for s in re.split(r"(\d+)", string_)]


def current_scope():
    """Current scope."""
    return current_graph()._scopes[-1]


def current_graph():
    """Current graph."""
    assert len(symjax._graphs)
    return symjax._graphs[-1]


class Graph(nx.DiGraph):
    def __init__(self, name, *args, **kwargs):

        super().__init__(*args, **kwargs)
        self._name = name
        self.reset()

    def reset(self):
        self.clear()
        self._current_scope = "/"
        self._scopes = [Scope("main", graph=self)]
        self._variables = {}
        self._placeholders = {}
        self._updates = {}
        self._ops = {}
        self._scopes_history = []

    def __repr__(self):
        msg = "Graph(name:{}, n_edges:{}, n_nodes:{})".format(
            self.name, self.n_edges, self.n_nodes
        )
        return msg

    def __str__(self):
        return self.__repr__()

    @property
    def n_nodes(self):
        return len(self.nodes)

    @property
    def n_edges(self):
        return len(self.edges)

    @property
    def name(self):
        return self._name

    def add(self, tensor, branch=None, **kwargs):

        if type(tensor) == dict:
            for i in tensor.values():
                self.add(i)
            self._updates.update(tensor)

        if not t.isvar(tensor):
            try:
                self.add_node(tensor, root=False)
                return tensor
            except TypeError:
                const_node = t.Constant(tensor)
                self.add_node(const_node, root=False)
                return const_node

        if tensor in self.nodes:
            return tensor

        self.scope.add(tensor)

        if isinstance(tensor, t.Placeholder) or isinstance(tensor, t.Variable):
            self.add_node(tensor, branch=branch, root=True)

        elif type(tensor) == t.OpTupleItem:
            self.add_node(tensor, root=False)
            self.add_edge(kwargs["parent"], tensor, index=kwargs["index"])

        elif isinstance(tensor, t.Op) or isinstance(tensor, t.OpTuple):

            if isinstance(tensor, t.RandomOp):
                self.add_node(
                    tensor,
                    branch=branch,
                    root=True,
                    jax_function=kwargs["jax_function"],
                    seed=kwargs["kwargs"]["_seed"],
                )
                del kwargs["kwargs"]["_seed"]
            else:
                self.add_node(
                    tensor,
                    branch=branch,
                    root=False,
                    jax_function=kwargs["jax_function"],
                )

            for i, arg in enumerate(kwargs["args"]):

                # all parents should already be in, however if one
                # of the arg is like a tuple then this is a new node
                # and thus it is not already in thus we add it
                node = self.add(arg)
                if self.has_edge(node, tensor):
                    self[node][tensor]["name"] += "+arg{}".format(i)
                else:
                    self.add_edge(node, tensor, name="arg{}".format(i))

            for key, arg in kwargs["kwargs"].items():

                node = self.add(arg)
                if self.has_edge(node, tensor):
                    self[node][tensor]["name"] += "+" + key
                else:
                    self.add_edge(node, tensor, name=key)

        elif type(tensor) == tuple:
            self.add_node(tensor, root=False)
            for i in tensor:
                other = self.add(i)
                self.add_edge(other, tensor)

        return tensor

    def roots(self, nodes, roots=None):

        if roots is None:
            roots = []

        if type(nodes) == tuple or type(nodes) == list:
            for node in nodes:
                self.roots(node, roots)
        else:
            if self.nodes[nodes]["root"]:
                roots.append(nodes)
            for i in nx.algorithms.ancestors(self, nodes):
                if self.nodes[i]["root"]:
                    roots.append(i)

        return list(set(roots))

    def clone(self, node, givens):
        names = [
            "{}{}->{}{}".format(m.scope, m.name, v.scope, v.name)
            for m, v in givens.items()
        ]
        names.sort()
        name = "_".join(names)
        branch = givens.copy()
        for start in givens.keys():
            for path in nx.all_simple_paths(self, source=start, target=node):
                for n in path[1:]:
                    if n in branch:
                        continue
                    args, kwargs = self._get_args_kwargs(n, evaluate=False)
                    args = [branch[arg] if arg in branch else arg for arg in args]
                    for arg in kwargs.keys():
                        if arg in branch:
                            kwargs[arg] = branch[arg]
                    fun = self.get_node_attribute(n, "jax_function")
                    kwargs.update(
                        {
                            "_jax_function": fun,
                            "_shape": n._shape,
                            "_dtype": n._dtype,
                            "name": n.name + "_clone",
                        }
                    )
                    branch[n] = t.Op(*args, **kwargs)
        return branch[node]

    def get_node_attribute(self, node, attr):
        return nx.get_node_attributes(self, attr)[node]

    def get(self, item, tracker=None, seed=None):
        """
        Example:

            >>> import symjax.tensor as T
            >>> w = T.ones(10).sum()+4
            >>> v = T.Variable(1., name='v', dtype='float32')
            >>> T.get(w+v)
            DeviceArray(15., dtype=float32)
        """

        if tracker is None:
            tracker = {}

        if not t.isvar(item):
            return item

        if type(item) == list:
            return [self.get(i, tracker) for i in item]
        elif type(item) == tuple:
            return tuple([self.get(i, tracker) for i in item])

        elif item in tracker:
            return tracker[item]

        elif isinstance(item, t.Placeholder):
            if item not in tracker:
                raise ValueError(" no value given for placeholder {}".format(item))

        elif isinstance(item, t.Variable) or type(item) == t.Constant:
            tracker[item] = item.value
            return tracker[item]

        elif type(item) == t.OpTupleItem:

            preds = list(self.predecessors(item))
            assert len(preds) == 1
            parent = preds[0]
            index = self.get_edge_data(parent, item)["index"]
            return self.get(parent, tracker)[index]

        elif isinstance(item, t.RandomOp):
            seed = seed or numpy.random.randint(0, 1000000)
            intra_seed = 1 or self.get_node_attribute(item, "seed")
            key = jax.random.PRNGKey(seed + intra_seed)
            args, kwargs = self._get_args_kwargs(item, tracker)
            tracker[item] = self.get_node_attribute(item, "jax_function")(
                key, *args, **kwargs
            )
            return tracker[item]

        elif isinstance(item, t.Op) or isinstance(item, t.OpTuple):

            # first get the actual parents nodes (aka inputs to the function)
            args, kwargs = self._get_args_kwargs(item, tracker)
            tracker[item] = self.get_node_attribute(item, "jax_function")(
                *args, **kwargs
            )

            return tracker[item]

        else:
            return item

    def _get_args_kwargs(self, node, tracker=None, evaluate=True):

        if evaluate:
            assert tracker is not None

            all_args = {
                self.get_edge_data(parent, node)["name"]: self.get(parent, tracker)
                for parent in self.predecessors(node)
            }
        else:
            all_args = {
                self.get_edge_data(parent, node)["name"]: parent
                for parent in self.predecessors(node)
            }
        # now we inspect if there are duplicate args
        for arg in list(all_args.keys()):
            if "+" in arg:
                items = arg.split("+")
                for item in items:
                    all_args.update({item: all_args[arg]})
                del all_args[arg]

        arg_names = [name for name in all_args.keys() if "arg" == name[:3]]
        arg_names = sorted(arg_names, key=natural_key)

        args = [all_args[name] for name in arg_names]
        for name in arg_names:
            del all_args[name]

        return args, all_args

    @property
    def scope(self):
        return self._scopes[-1]

    def variables(self, trainable=True):
        variables = [n for n in self.nodes if isinstance(n, t.Variable)]
        if trainable is None:
            return variables
        return [v for v in variables if v.trainable == trainable]

    @property
    def placeholders(self):
        placeholders = [n for n in self.nodes if type(n) == t.Placeholder]
        return placeholders

    @property
    def ops(self):
        ops = [n for n in self.nodes if type(n) in [t.Op, t.OpTupleItem]]
        return ops

    def updates(self, name="*"):
        sub = {}
        for v in self._updates:
            matched = fnmatch.filter([v.name], name)
            if len(matched):
                sub[v] = self._updates[v]
        return sub


class Scope:
    """
    Defining scope for any variable/operation to be in.

    Example
    -------

    .. doctest::
        >>> import symjax
        >>> import symjax.tensor as T
        >>> v = T.Variable(1, name='v')
        >>> # the current (default) scope is the root of the graph
        >>> print(v.scope)
        /
        >>> with symjax.Scope('my_scope'):
        ...     w = T.Variable(2, name='w')
        ...     out = v*w
        >>> print(out.scope)
        /my_scope/
        >>> print(w.scope)
        /my_scope/
        >>> #it is also possible to keep a scope persistently
        >>> scope = symjax.Scope('second_scope')
        >>> with scope:
        ...     other = out * w
        >>> print(other.scope)
        /second_scope/
        >>> # this allows to keep track directly of internal ops
        >>> print(scope.ops)
        [Op(name=multiply, fn=multiply, shape=(), dtype=int32, scope=/second_scope/)]


    """

    def __init__(self, name, graph=None, reattach=False):
        """Constructor."""
        if graph is None:
            graph = current_graph()
        self.graph = graph
        assert len(name)
        assert "_" != name[-1]
        self.name = name
        self.full_name = None

    def __enter__(self):
        """Set global variables."""

        current = self.graph._current_scope
        cpt = 0
        if current + self.name + "/" in self.graph._scopes_history:
            while (
                current + self.name + "_{}/".format(cpt) in self.graph._scopes_history
            ):
                cpt += 1
            self.name += "_{}".format(cpt)

        self.graph._current_scope += self.name + "/"
        self.full_name = self.graph._current_scope
        self.graph._scopes.append(self)
        self.graph._scopes_history.append(self.full_name)

    def __exit__(self, *a):
        """Delete globals."""
        self.graph._scopes.pop(-1)
        if self.graph._scopes[-1].full_name is not None:
            self.graph._current_scope = self.graph._scopes[-1].full_name
        else:
            self.graph._current_scope = "/"

    def save_variables(self, path):
        """Save graph."""
        numpy.savez(
            path, **dict([(v.name, symjax.tensor.get(v)) for v in self.variables])
        )

    @property
    def variables(self):
        return get_variables(self.full_name + "*")

    @property
    def ops(self):
        return get_ops(self.full_name + "*")

    @property
    def placeholders(self):
        return get_placeholders(self.full_name + "*")

    def load_variables(self, path):
        """Load graph."""

        data = numpy.load(path)
        for name, value in data.items():
            self.variable[name].update(value)

    def reset(self):

        for var in self.variables:
            var.reset()

    def add(self, tensor):

        if not t.isvar(tensor) or type(tensor) == tuple:
            return

        # fake graph entrance if not used by user
        if self.full_name is None:
            self.__enter__()

        # in this case we were given updates
        if type(tensor) == dict:
            self.graph._updates.update(tensor)
            return

        tensor.scope = self.full_name
        name = self.full_name + tensor.name
        if isinstance(tensor, symjax.tensor.Placeholder):
            names = self.graph._placeholders
        elif isinstance(tensor, symjax.tensor.Variable):
            names = self.graph._variables
        else:
            names = self.graph._ops

        if name not in names.keys():
            names[name] = tensor
            return

        count = 1
        while True:
            if name + "_" + str(count) in names.keys():
                count += 1
            else:
                break
        names[name + "_" + str(count)] = tensor
        tensor._set_name(tensor.name + "_" + str(count))


def reset_variables(name="*", trainable=None):
    """
    utility to reset variables based on their names

    Parameters
    ----------

    name: str (default=*)
        the name (or part of the name) of all the variables that should be
        reset, it can include the glob (*) searching for all matching
        names

    trainable: bool or None (optional, default=None)
        is not None, it will only reset from the matched variables the ones that
        trainable attribute matches the given one


    Returns
    -------

    None

    Example
    -------

    .. doctest::

        >>> import symjax
        >>> w = symjax.tensor.Variable(1., name='w', dtype='float32')
        >>> x = symjax.tensor.Variable(2., name='x', dtype='float32')
        >>> f = symjax.function(outputs=[w, x], updates={w:w + 1,x:x + 1})
        >>> for i in range(10):
        ...    print(f())
        [array(1., dtype=float32), array(2., dtype=float32)]
        [array(2., dtype=float32), array(3., dtype=float32)]
        [array(3., dtype=float32), array(4., dtype=float32)]
        [array(4., dtype=float32), array(5., dtype=float32)]
        [array(5., dtype=float32), array(6., dtype=float32)]
        [array(6., dtype=float32), array(7., dtype=float32)]
        [array(7., dtype=float32), array(8., dtype=float32)]
        [array(8., dtype=float32), array(9., dtype=float32)]
        [array(9., dtype=float32), array(10., dtype=float32)]
        [array(10., dtype=float32), array(11., dtype=float32)]
        >>> # reset only the w variable
        >>> symjax.reset_variables('w')
        >>> # reset all variables
        >>> symjax.reset_variables('*')

    """

    variables = get_variables(name, trainable)
    for var in variables:
        var.reset()


def save_variables(name, path):
    """Save graph."""
    matched = fnmatch.filter(symjax._variables.keys(), name)
    numpy.savez(
        path,
        **dict(
            [
                (
                    symjax._variables[v].scope + symjax._variables[v].name,
                    symjax.tensor.get(symjax._variables[v]),
                )
                for v in matched
            ]
        )
    )


def load_variables(name, path_or_file, scope_mapping=None):
    """Load graph."""

    if type(path_or_file) == str:
        if path_or_file[-4:] != ".npz":
            path_or_file += ".npz"

    scope_mapping = scope_mapping or {}

    matched = fnmatch.filter(symjax._variables.keys(), name)
    data = numpy.load(path_or_file)
    for name in matched:
        if symjax._variables[name].scope in scope_mapping:
            name_in_file = (
                scope_mapping[symjax._variables[name].scope]
                + "/"
                + symjax._variables[name].name
            )
        else:
            name_in_file = name
        if name_in_file not in data:
            raise Warning("{} not in loaded file".format(name_in_file))
        symjax._variables[name].update(data[name_in_file])


def get_variables(name="*", trainable=None):
    matched = current_graph().variables(trainable)
    output = []
    for m in matched:
        if len(fnmatch.filter([m.name], name)):
            output.append(m)
    return output


def get_placeholders(name="*"):
    """
    Same as symjax.variable but for placeholders
    """
    matched = current_graph().placeholders
    output = []
    for m in matched:
        if len(fnmatch.filter([m.name], name)):
            output.append(m)
    return output


def get_ops(name="*"):
    """
    Same as symjax.variable but for ops
    """
    matched = current_graph().ops
    output = []
    for m in matched:
        if len(fnmatch.filter([m.scope + m.name], name)):
            output.append(m)
    return output


def add_updates(udpates):
    current_scope().add_updates(updates)


def get_updates(name="*"):
    """
    Same as symjax.get_variables but for updates
    """
    return current_graph().updates(name)


def gradients(scalar, variables):
    """Compute the gradients of a scalar w.r.t to a given list of variables.

    Arguments
    ---------
    scalar: :class:`symjax.tensor.base.Tensor`
        the variable to differentiate

    variables: List or Tuple
        the variables used to compute the derivative.

    Returns
    -------

        gradients: Tuple
            the sequency of gradients ordered as given in the input variables

    Example
    -------

    .. doctest::

        >>> import symjax
        >>> w = symjax.tensor.ones(3)
        >>> x = symjax.tensor.Variable(2., name='x', dtype='float32')
        >>> l = (w ** 2).sum() * x
        >>> g = symjax.gradients(l, [w])[0]
        >>> f = symjax.function(outputs=g, updates={x:x + 1})
        >>> for i in range(2):
        ...    print(f())
        [4. 4. 4.]
        [6. 6. 6.]

    """
    if numpy.prod(scalar.shape) != 1:
        raise RuntimeError("the variable to differentiate is not a scalar")
    if not isinstance(scalar, t.Tensor):
        raise RuntimeError("the variable used in gradients should be a Tensor type")

    if scalar.shape != ():
        scalar = scalar.sum()
    if isinstance(variables, t.Tensor):
        input_variables = [variables]
        input_list = False
    else:
        input_variables = variables
        input_list = True

    # get the argnum of the variables that we differentiate one
    argnums = list(range(len(input_variables)))

    # get all the roots of the scalar, this is needed as otherwise they are not
    # as the input of the gradient function and thus a change of
    # their value will not change the gradient computation, we also ensure
    # uniqueness
    input_variables += [
        i for i in current_graph().roots(scalar) if i not in input_variables
    ]

    # create a dummy function that is needed for jax to compute a gradient func
    # this function is the one that builds the graph of computation from all
    # roots
    # to the scalar varible s.t. automatic diffenrentiation can be applied

    def fn(*args):
        return current_graph().get(scalar, dict(zip(input_variables, list(args))))

    # now we obtain the grad function. In fact, Jax returns a function that,
    # when it is called, returns the gradient values, this function is then
    # used to generate the Tuple of symbolic variables
    grad_fn = jax.grad(fn, argnums)
    wrap_fn = t.jax_wrap(grad_fn)
    if input_list:
        return wrap_fn(*input_variables)
    else:
        return wrap_fn(*input_variables)[0]


def jacobians(tensor, variables, mode="forward"):
    """Compute the jacobians of a tensor w.r.t to a given list of variables.

    The tensor needs not to be a vector, but will be treated as such. For
    example if tensor.shape is (10, 3, 3) and a variable shape if (10, 10)
    the resulting jacobian has shape (10, 3, 3, 10, 10). It is possible
    to specify the mode forward or backward. For tall jacobians, forward
    is faster and vice-versa.

    Arguments
    ---------

        vector: Tensor
            the variable to differentiate

        variables: List or Tuple
            the variables used to compute the derivative.

    Returns
    -------

        jacobians: Tuple
            the sequency of gradients ordered as given in the input variables
            :param tensor:
            :param mode:
    """
    # get all the roots of the scalar, this is needed as otherwise they are not
    # as the input of the gradient function and thus a change of
    # their value will not change the gradient computation, we also ensure
    # uniqueness
    all_roots = list(set(tensor.roots + variables))

    # get the argnum of the variables that we differentiate one
    argnums = [all_roots.index(var) for var in variables]

    # create a dummy function that is needed for jax to compute a gradient func
    # this function is the one that builds the graph of computation from
    # all roots
    # to the scalar varible s.t. automatic diffenrentiation can be applied
    def fn(*args):
        return symjax.tensor.get(tensor, dict(zip(all_roots, list(args))))

    # now we obtain the jacobian function. In fact, Jax returns a function that
    # when it is called, returns the jacobian values, this function is then
    # used to generate the Tuple of symbolic variables
    if mode == "forward":
        jacob_fn = jacfwd(fn, argnums)
    elif mode == "backward":
        jacob_fn = jacrev(fn, argnums)
    else:
        raise RuntimeError(
            "mode {} not recognized, use forward or backward".format(mode)
        )
    wrap_fn = t.jax_wrap(jacob_fn, False)
    return wrap_fn(*all_roots)


class function:
    """Generate a user function that compiles a computational graph.

    Based on given inputs, outputs and update policy of variables. This
    function internally jit compile the underlying jax computational
    graph for performances.

    Arguments
    ---------

        classargs: trailing tuple
            the inputs to the function to be compiled. The tuple should
            contain all the placeholders that are roots of any output
            given of the function and update values

        outputs: List (optional)
            the outputs of the function, if a single element, it can be
            given as a standalone and not a list

        updates: Dict (optional)
            the dictionnary of updates as per {var:new_value} for any
            variable of the graph

        backend: 'cpu' or 'gpu'
            the backend to use to run the function on

        default_value: not implemented
            not implemented

    Returns
    -------

        callable:
            the user frontend function that takes the specified inputs,
            returns the specified outputs and perform internally the
            updates

    Examples
    --------

        >>> import symjax
        >>> import symjax.tensor as T
        >>> x = T.ones((4, 4))
        >>> xs = x.sum() + 1
        >>> f = symjax.function(outputs=xs)
        >>> print(f())
        17.0

        >>> w = T.Variable(0., name='w', dtype='float32')
        >>> increment = symjax.function(updates={w: w + 1})
        >>> for i in range(10):
        ...     increment()
        >>> print(w.value)
        10.0

    """

    def __init__(
        self,
        *classargs,
        outputs=[],
        updates=None,  # noqa
        device=None,
        backend=None,
        default_value=None
    ):
        """Initialize."""
        # check the given updates (if any) and ensure that they only
        # update Variable objects
        if updates is None:
            updates = {}
        else:
            current_graph().add(updates)

        for update in updates.keys():
            if not isinstance(update, t.Variable):
                raise RuntimeError(
                    "{} is not a Variable and cannot be updated".format(update)
                )

        # ensure that all inputs are actual placeholders or variables
        for arg in classargs:
            if not isinstance(arg, t.Tensor):
                raise RuntimeError(
                    "{} is not a Tensor type. Only tensor types can be"
                    + "function inputs".format(arg)
                )

        # gather all roots, they need to be explicit as inputs of the
        # underlying functions otherwise they are treated as constants
        # and any change in their value will not appear when running the
        # function
        self.updates_keys = list(updates.keys())
        self.updates_values = list(updates.values())
        self.classargs = classargs
        self.outputs = outputs

        outs = self.updates_values
        outs += [outputs] if isinstance(outputs, t.Tensor) else outputs

        self.all_roots = set(symjax.current_graph().roots(outs))

        # check the function inputs, they must be at least contain all the
        # placeholders needed to compute the outputs values
        placeholders_in_root = [
            x for x in self.all_roots if isinstance(x, t.Placeholder)
        ]

        # check for
        non_givens = set(placeholders_in_root) - set(self.classargs)
        if len(non_givens) > 0:
            raise RuntimeError(
                "Missing placeholders form the function inputs: {}".format(non_givens)
            )

        # the roots are made of variables, random tensors, placeholders. We
        # already ensured that all placeholders are given as inputs to the
        # function. Now we must ensure that the other ones will also be given
        # as inputs to not be treated as constants by jax.
        # we also remove update keys because we will expicitly feed them
        self.extra_inputs = set(self.all_roots) - set(self.classargs).union(
            self.updates_keys
        )

        self.extra_inputs = list(self.extra_inputs)
        allargs = list(self.classargs) + self.updates_keys + self.extra_inputs

        def to_jit(*jitargs, seed):

            feed_dict = dict(zip(allargs, jitargs))
            feed_dict.update({"rng": seed})
            outputs = [self.outputs, self.updates_values]
            return symjax.current_graph().get(outputs, feed_dict)

        # take care of the presence of -1 dimensions
        to_vectorize = [0 if 0 in a.shape else None for a in allargs]

        if any(to_vectorize):
            self.jited = jax.jit(
                jax.vmap(to_jit, to_vectorize), device=device, backend=backend
            )
        else:
            # we compile our underlying function using jit for performances
            self.jited = jax.jit(to_jit, device=device, backend=backend)

        # define the frontend function that takes as input the inputs variables
        # and internally compute and update the variables from updates if any
        def meta(*fnargs, rng):
            # ensure that the number of arguments is correct
            assert len(fnargs) == len(self.classargs)
            for fnarg, classarg in zip(fnargs, self.classargs):
                if hasattr(fnarg, "shape"):
                    if fnarg.shape != classarg.shape and 0 not in classarg.shape:
                        raise RuntimeError(
                            "wrong input given for {}".format(classarg)
                            + ", given is {}".format(fnarg)
                            + ", shape={}".format(fnarg.shape)
                        )

            # retreive the function outputs, updated values and apply them
            jited_add_inputs = symjax.current_graph().get(
                self.updates_keys + self.extra_inputs, tracker={"rng": rng}
            )
            jitoutputs, jitupdates = self.jited(*fnargs, *jited_add_inputs, seed=rng)
            for key, update in zip(self.updates_keys, jitupdates):
                key.update(update)

            if type(jitoutputs) == list or type(jitoutputs) == tuple:
                npy_jitoutputs = [
                    jax.api.device_get(arr)
                    if isinstance(arr, jax.interpreters.xla.DeviceArray)
                    else arr
                    for arr in jitoutputs
                ]
                return npy_jitoutputs

            else:
                return jax.api.device_get(jitoutputs)

        self.meta = meta

    def __call__(self, *args, rng=None):
        """Callable fn."""
        # in the presence of RandomTensor(s) in the graph, we keep track of the
        # number of functions calls to keep accumulating the PRNGKey of the jax
        # key, otherwise each function call returns the same realisation

        if rng is None:
            rng = random._seed
            random._seed += 1
        pargs = [numpy.array(arg) if type(arg) == list else arg for arg in args]
        outputs = self.meta(*pargs, rng=rng)
        if type(outputs) == list:
            if len(outputs) == 0:
                return None
        return outputs



import os
import pickle, gzip
import urllib.request
import numpy as np
import time
import zipfile
from scipy.io import arff
import io
from tqdm import tqdm


class univariate_timeseries:
    """univariate_timeseries
    """

    @staticmethod
    def download(path):
        """
        Download the univariate_timeseries dataset
        path

        Parameters
        ----------

            path: str
                the path where the downloaded files will be stored. If the
                directory does not exist, it is created.
        """

        # Check if directory exists
        if not os.path.isdir(path + "univariate_timeseries"):
            print("Creating univariate_timeseries Directory")
            os.mkdir(path + "univariate_timeseries")

        # Check if file exists
        if not os.path.exists(path + "univariate_timeseries/Univariate2018_arff.zip"):
            td = time.time()
            print("Creating univariate timeseries")
            url = "http://www.timeseriesclassification.com/Downloads/Archives/Univariate2018_arff.zip"
            urllib.request.urlretrieve(
                url, path + "univariate_timeseries/Univariate2018_arff.zip"
            )

    @staticmethod
    def load(path=None):
        """
        Parameters
        ----------
            path: str (optional)
                default ($DATASET_PATH), the path to look for the data and
                where the data will be downloaded if not present

        Returns
        -------

            train_images: array

            train_labels: array

            valid_images: array

            valid_labels: array

            test_images: array

            test_labels: array

        """

        if path is None:
            path = os.environ["DATASET_PATH"]

        univariate_timeseries.download(path)

        t0 = time.time()

        # Loading the fileunivariate_timeseries
        f = zipfile.ZipFile(path + "univariate_timeseries/Univariate2018_arff.zip")
        # init. the data array
        all_data = {}
        for name in tqdm(f.namelist()):
            if "TRAIN.txt" in name or "TEST.txt" in name:
                file = f.read(name)
                if "PLAID" in name:
                    data = np.loadtxt(io.BytesIO(file), delimiter=",")
                else:
                    data = np.loadtxt(io.BytesIO(file))

                dataset = name[:-4].split("/")[-2]
                part = "TRAIN" if "TRAIN" in name else "TEST"
                all_data[dataset + "/" + part] = data
        return all_data



import os
import gzip
import urllib.request
import numpy as np
import time
import zipfile
import io
from scipy.io.wavfile import read as wav_read
from tqdm import tqdm


class freefield1010:
    """Audio binary classification, presence or absence of bird songs.
    `freefield1010 <http://machine-listening.eecs.qmul.ac.uk/bird-audio-detection-challenge/#downloads>`_. 
    is a collection of over 7,000 excerpts from field recordings 
    around the world, gathered by the FreeSound project, and then standardised 
    for research. This collection is very diverse in location and environment, 
    and for the BAD Challenge we have newly annotated it for the 
    presence/absence of birds.
    """

    def download(path):

        if path is None:
            path = os.environ["DATASET_path"]

        # Load the dataset (download if necessary) and set
        # the class attributes.

        print("Downloading freefield1010")
        t = time.time()

        if not os.path.isdir(path + "freefield1010"):
            print("\tCreating Directory")
            os.mkdir(path + "freefield1010")

        if not os.path.exists(path + "freefield1010/ff1010bird_wav.zip"):
            url = "https://archive.org/download/ff1010bird/ff1010bird_wav.zip"
            urllib.request.urlretrieve(url, path + "freefield1010/ff1010bird_wav.zip")

        if not os.path.exists(path + "freefield1010/ff1010bird_metadata.csv"):
            url = "https://ndownloader.figshare.com/files/6035814"
            urllib.request.urlretrieve(
                url, path + "freefield1010/ff1010bird_metadata.csv"
            )

        print("freefield1010 downloaded in {} sec.".format(time.time() - t))

    def load(path=None):
        if path is None:
            path = os.environ["DATASET_path"]

        freefield1010.download(path)
        t = time.time()

        # load labels
        labels = np.loadtxt(
            path + "freefield1010/ff1010bird_metadata.csv",
            delimiter=",",
            skiprows=1,
            dtype="int32",
        )
        # load wavs
        f = zipfile.ZipFile(path + "freefield1010/ff1010bird_wav.zip")
        # init. the data array
        N = labels.shape[0]
        wavs = np.empty((N, 441000 // subsample), dtype="float32")
        for i, files_ in tqdm(enumerate(labels[:, 0]), ascii=True, total=N):
            wavfile = f.read("wav/" + str(files_) + ".wav")
            byt = io.BytesIO(wavfile)
            wavs[i] = wav_read(byt)[1].astype("float32")[::subsample]

        labels = labels[:, 1]

        return wavs, labels



import io
import os
import time
import urllib.request
import zipfile

import numpy as np
from scipy.io.wavfile import read as wav_read
from tqdm import tqdm


class dclde:
    """
    The high-frequency dataset consists of marked encounters with echolocation
    clicks of species commonly found along the US Atlantic Coast, and in the
    Gulf of Mexico:

    Mesoplodon europaeus - Gervais' beaked whale
    Ziphius cavirostris - Cuvier's beaked whale
    Mesoplodon bidens - Sowerby's beaked whale
    Lagenorhynchus acutus - Atlantic white-sided dolphin
    Grampus griseus - Risso's dolphin
    Globicephala macrorhynchus - Short-finned pilot whale
    Stenella sp. - Stenellid dolphins
    Delphinid type A
    Delphinid type B
    Unidentified delphinid - delphinid other than those described above

    The goal for these datasets is to identify acoustic encounters by species
    during times when animals were echolocating. Analysts examined data for
    echolocation clicks and approximated the start and end times of acoustic
    encounters. Any period that was separated from another one by five minutes
    or more was marked as a separate encounter. Whistle activity was not
    considered. Consequently, while the use of whistle information during
    echolocation activity is appropriate, reporting a species based on whistles
    in the absence of echolocation activity will be considered a false positive
    for this classification task.
    """

    def download(path):
        """ToDo
        """

        # Load the dataset (download if necessary) and set
        # the class attributes.

        print("Loading DCLDE")
        t = time.time()

        if not os.path.isdir(path + "DCLDE"):
            print("\tCreating Directory")
            os.mkdir(path + "DCLDE")
        if not os.path.exists(path + "DCLDE/DCLDE_LF_Dev.zip"):
            url = "http://sabiod.univ-tln.fr/workspace/DCLDE2018/DCLDE_LF_Dev.zip"
            with DownloadProgressBar(
                unit="B", unit_scale=True, miniters=1, desc="Wav files"
            ) as t:
                urllib.request.urlretrieve(url, path + "DCLDE/DCLDE_LF_Dev.zip")

    def load(window_size=441000, path=None):
        """ToDo
        """
        if path is None:
            path = os.environ["DATASET_path"]
        dclde.download(path)

        # Loading the files
        f = zipfile.ZipFile(path + "DCLDE/DCLDE_LF_Dev.zip")
        wavs = list()
        #    labels  = list()
        for zipf in tqdm(f.filelist, ascii=True):
            if ".wav" in zipf.filename and ".d100." in zipf.filename:
                wavfile = f.read(zipf)
                byt = io.BytesIO(wavfile)
                wav = wav_read(byt)[1].astype("float32")
                for s in range(len(wav) // window_size):
                    wavs.append(wav[s * window_size : (s + 1) * window_size])
        #            labels.append(zipf.filename.split('/')[2])
        #    return wavs,labels
        wavs = np.expand_dims(np.asarray(wavs), 1)
        dataset.add_variable({"signals": {"train_set": wavs}})

        print(
            "Dataset freefield1010 loaded in", "{0:.2f}".format(time.time() - t), "s."
        )
        return dataset



import numpy as np
from multiprocessing import Pool, Queue, Lock, Process
from scipy import ndimage


def create_cmap(values, colors):

    from matplotlib.pyplot import Normalize
    import matplotlib

    norm = Normalize(min(values), max(values))
    tuples = list(zip(map(norm, values), colors))
    cmap = matplotlib.colors.LinearSegmentedColormap.from_list("", tuples)
    return cmap, norm


def patchify_1d(x, window_length, stride):
    """extract patches from a numpy array

    Parameters
    ----------

    x: array-like
        the input data to extract patches from, any shape, the last dimension
        is the one being patched

    window_length: int
        the length of the patches

    stride: int
        the amount of stride (bins separating two consecutive patches

    Returns
    -------

    x_patches: array-like
        the number of patches is put in the pre-last dimension (-2)
    """

    n_windows = (x.shape[-1] - window_length) // stride + 1
    new_x = np.empty(x.shape[:-1] + (n_windows, window_length))
    for n in range(n_windows):
        new_x[..., n, :] = x[..., n * stride : n * stride + window_length]
    return new_x


def patchify_2d(x, window_length, stride):
    # TODO
    return None


def train_test_split(*args, train_size=0.8, stratify=None, seed=None):
    """split given data into two non overlapping sets

    Parameters
    ----------

    *args: inputs
        the sets to be split by the function

    train_size: scalar
        the amount of data to put in the first set, either an integer value
        being the actual number of data to keep, or a ratio (0 to 1 number)

    stratify: array (optional)
        the optimal stratify guide to spit the array s.t. the same proportion
        based on the stratify array is kep in both set based on the proportion
        of the split

    seed: integer (optional)
        the seed for the random number generator for reproducibility

    Returns
    -------

    train_set: list
        returns the train data, the list has the members of *args split

    test_set: list
        returns the test data, the list has the members of *args split

    Example
    -------

    .. code-block:: python

       x = numpy.random.randn(100, 4)
       y = numpy.random.randn(100)

       train, test = train_test_split(x, y, train_size=0.5)
       print(train[0].shape, train[1].shape)
       # (50, 4) (50,)
       print(test[0].shape, test[1].shape)
       # (50, 4) (50,)


    """
    if stratify is not None:
        train_indices = list()
        test_indices = list()
        for c in set(list(stratify)):
            c_indices = np.where(stratify == c)[0]
            np.random.RandomState(seed=seed).shuffle(c_indices)
            if train_size > 1:
                cutoff = train_size
            else:
                cutoff = int(len(c_indices) * train_size)
            train_indices.append(c_indices[:cutoff])
            test_indices.append(c_indices[cutoff:])
        train_indices = np.concatenate(train_indices, 0)
        test_indices = np.concatenate(test_indices, 0)
    else:
        indices = np.random.RandomState(seed=seed).permutation(len(args[0]))
        if train_size > 1:
            assert type(train_size) == int
            cutoff = train_size
        else:
            cutoff = int(len(args[0]) * train_size)
        print(cutoff)
        train_indices = indices[:cutoff]
        test_indices = indices[cutoff:]
    train_set = [arg[train_indices] for arg in args]
    test_set = [arg[test_indices] for arg in args]
    if len(args) == 1:
        return train_set[0], test_set[0]
    return train_set, test_set


class batchify:
    def __init__(
        self,
        *args,
        batch_size,
        option="random",
        load_func=None,
        extra_process=0,
        n_batches=None
    ):
        """

        Parameters
        ----------

        load_func: None or list of func
            same length as the number of args. A function is called on a single
            datum and it can be used to apply some normalization but its main
            goal is to load files if the args were list of filenames

        extra_processes: int (optional)
            if there is no load_func then extra process is useless

        n_batches: int (optional)
            the number of batches to produce, only used if option is random, if
            not given it is taken to be the length of the data divided by the
            batch_size

        Returns
        -------

        *batch_args: list
            the iterator containing the batch values
            of each arg in args

        Example
        -------

        .. code-block:: python

        for x, y in batchify(X, Y):
            train(x, y)

        """

        self.n_batches = n_batches or len(args[0]) // batch_size
        self.args = args
        self.start_index = 0
        self.option = option
        self.batch_size = batch_size
        self.extra_process = extra_process
        self.terminate = False

        if option == "random_see_all":
            self.permutation = np.random.permutation(len(args[0]))

        # set up load function
        if load_func is None:
            self.load_func = (None,) * len(args)
        else:
            self.load_func = []
            for i in range(len(load_func)):
                if load_func[i] is None:
                    self.load_func.append(None)
                elif extra_process == 0:

                    def fn(args, queue, f=load_func[i]):
                        result = np.asarray([f(arg) for arg in args])
                        queue.put(result)

                    self.load_func.append(fn)
                else:

                    def fn(lock, data, queue, f=load_func[i]):
                        lock.acquire()
                        with Pool(processes=extra_process) as pool:
                            result = pool.map(f, data)
                        queue.put(np.asarray(result))
                        lock.release()

                    self.load_func.append(fn)
        assert np.prod([len(args[0]) == len(arg) for arg in args[1:]])

        self.queues = [Queue() for f in self.load_func]
        self.locks = [Lock() for f in self.load_func]

        # launch the first batch straight away
        batch = self.get_batch()
        self.launch_process(batch)

    def chunk(self, items, n):
        for i in range(0, len(items), n):
            yield items[i : i + n]

    def launch_process(self, batch):

        for b, lock, load_func, queue in zip(
            batch, self.locks, self.load_func, self.queues
        ):
            if load_func is None:
                queue.put(b)
            elif load_func is not None and self.extra_process == 0:
                load_func(b, queue)
            elif load_func is not None and self.extra_process > 0:
                # first chunk for each process and launch a process
                p = Process(target=load_func, args=(lock, b, queue))
                p.start()

    def __iter__(self):
        return self

    def get_batch(self):

        indices = (self.start_index, self.start_index + self.batch_size)

        # check if we exhausted the samples
        if self.option == "random":
            if indices[1] > self.batch_size * self.n_batches:
                raise StopIteration()
        elif indices[1] > len(self.args[0]):
            raise StopIteration()

        # proceed to get the data
        if self.option == "random_see_all":
            perm = self.permutation[indices[0] : indices[1]]
            batch = [
                arg[perm] if hasattr(arg, "shape") else [arg[i] for i in perm]
                for arg in self.args
            ]
        elif self.option == "continuous":
            batch = [arg[indices[0] : indices[1]] for arg in self.args]
        elif self.option == "random":
            perm = np.random.randint(0, len(self.args[0]), self.batch_size)
            batch = [
                arg[perm] if hasattr(arg, "shape") else [arg[i] for i in perm]
                for arg in self.args
            ]
        return batch

    def __next__(self):

        if self.terminate:
            raise StopIteration()

        # we prepare the next batch if possible
        try:
            self.start_index += self.batch_size
            batch = self.get_batch()
            self.launch_process(batch)
        except StopIteration:
            self.terminate = True

        if len(self.queues) == 1:
            return self.queues[0].get()
        else:
            return tuple([queue.get() for queue in self.queues])

    def apply_func(self, batch):
        # now check if needs to apply a load func
        for i, load_func in enumerate(self.load_func):
            if load_func is not None:
                batch[i] = load_func(batch[i])


def vq_to_boundary(states, N, M, duplicate=1):
    # the following should take as input a collection of points
    # in the input space and return a list of binary states, each
    # element of the list is for 1 specific layer and it is a 2D array
    if states.ndim > 1:
        states = vq_to_values(states)
    partitioning = values_to_boundary(
        states.reshape((N, M)).astype("float32"), duplicate
    )
    return partitioning


def values_to_boundary(x, duplicate=0):
    # compute each directional (one step) derivative as a boolean mask
    # representing jump from one region to another and add them (boolean still)
    g_vertical = np.greater(
        np.pad(np.abs(x[1:] - x[:-1]), ((1, 0), (0, 0)), "constant"), 0
    )
    g_horizontal = np.greater(
        np.pad(np.abs(x[:, 1:] - x[:, :-1]), [[0, 0], [1, 0]], "constant"), 0
    )
    g_diagonaldo = np.greater(
        np.pad(np.abs(x[1:, 1:] - x[:-1, :-1]), [[1, 0], [1, 0]], "constant"), 0,
    )
    g_diagonalup = np.greater(
        np.pad(np.abs(x[:-1:, 1:] - x[1:, :-1]), [[1, 0], [1, 0]], "constant"), 0,
    )
    overall = g_vertical + g_horizontal + g_diagonaldo + g_diagonalup
    if duplicate > 0:
        overall = np.stack(
            [np.roll(overall, k, 1) for k in range(duplicate + 1)]
            + [np.roll(overall, k, 0) for k in range(duplicate + 1)]
        ).sum(0)
        overall[:duplicate] *= 0
        overall[-duplicate:] *= 0
        overall[:, :duplicate] *= 0
        overall[:, -duplicate:] *= 0
    return np.greater(overall, 0).astype("float32")


def vq_to_values(states, vq_to_value_dict=None, return_dict=False):
    """Given binary masks obtained for example from the ReLU activation, 
    convert the binary vectors to a single float scalar, the same scalar for
    the same masks. This allows to drastically reduce the memory overhead to
    keep track of a large number of masks. The mapping mask -> real is kept
    inside a dict and thus allow to go from one to another. Thus given a large
    collection of masks, one ends up with a large collection of scalars and
    a mapping mask <-> real only for unique masks and thus allow reduced memory
    requirements.

    Parameters
    ----------

    states : bool matrix
        the matrix of shape (#samples,#binary values). Thus if using a deep net
        the masks of ReLU for all the layers must first be flattened prior
        using this function.

    vq_to_value_dict : dict
        optional dict containing an already built mask <-> real mapping which
        should be used and updated given the states value.

    return_dict : bool
        if the update/created dict should be return as part of the output

    Returns
    -------

    values : scalar vector
        a vector of length #samples in which each entry is the scalar
        representation of the corresponding mask from states

    vq_to_value_dict : dict (optional)
        the newly created or updated dict mapping mask to value and vice-versa.


    """
    if vq_to_value_dict is None:
        vq_to_value_dict = dict()
    if "count" not in vq_to_value_dict:
        vq_to_value_dict["count"] = 0
    values = np.zeros(states.shape[0])
    for i, state in enumerate(states):
        str_s = "".join(state.astype("uint8").astype("str"))
        if str_s not in vq_to_value_dict:
            vq_to_value_dict[str_s] = vq_to_value_dict["count"]
            vq_to_value_dict["count"] += 0.001
        values[i] = vq_to_value_dict[str_s]
    return values


def resample_images(
    images,
    target_shape,
    ratio="same",
    order=1,
    mode="nearest",
    data_format="channels_first",
):

    if data_format == "channels_first":
        output_images = np.zeros(
            (len(images), images[0].shape[0]) + target_shape, dtype=images[0].dtype,
        )
    else:
        output_images = np.zeros(
            (len(images),) + target_shape + (images[0].shape[-1],),
            dtype=images[0].dtype,
        )

    for i, image in enumerate(images):

        # the first step is to resample to fit it into the target shape
        if data_format != "channels_first":
            image = image.transpose(2, 0, 1)

        if ratio == "same":
            width_change = target_shape[1] / image.shape[2]
            height_change = target_shape[0] / image.shape[1]
            change = min(width_change, height_change)
        x = np.linspace(0, image.shape[1] - 1, int(image.shape[1] * change))
        y = np.linspace(0, image.shape[2] - 1, int(image.shape[2] * change))
        coordinates = np.stack(np.meshgrid(x, y))
        coordinates = np.stack([coordinates[0].reshape(-1), coordinates[1].reshape(-1)])

        new_image = np.stack(
            [
                ndimage.map_coordinates(channel, coordinates, order=order, mode=mode)
                .reshape((len(y), len(x)))
                .T
                for channel in image
            ]
        )
        # now we position this image into the output
        middle_height = (target_shape[0] - len(x)) // 2
        middle_width = (target_shape[1] - len(y)) // 2
        if data_format != "channels_first":
            output_images[
                i,
                middle_height : middle_height + len(x),
                middle_width : middle_width + len(y),
                :,
            ] = new_image.transpose(1, 2, 0)
        else:
            output_images[
                i,
                :,
                middle_height : middle_height + len(x),
                middle_width : middle_width + len(y),
            ] = new_image
    return output_images



import inspect
import sys

import jax
import jax.lax as jla
import numpy

from symjax.tensor import jax_wrap

NAMES = [c[0] for c in inspect.getmembers(jax.nn, callable)]
module = sys.modules[__name__]
for name in NAMES:
    if name == "one_hot":
        continue
    module.__dict__.update({name: jax_wrap(jax.nn.__dict__[name])})


def log_1_minus_sigmoid(x):
    return -module.__dict__["softplus"](x)


@jax_wrap
def convNd(
    input,
    filter,
    strides=1,
    padding="VALID",
    input_format=None,
    filter_format=None,
    output_format=None,
    input_dilation=None,
    filter_dilation=None,
):
    """General n-dimensional convolution operator, with input/filter dilation.

    Wraps Jax's conv_general_dilated functin, and thus also the XLA's `Conv
    <https://www.tensorflow.org/xla/operation_semantics#conv_convolution>`_
    operator.

    Args:
        input (Tensor): a rank `n+2` dimensional input array.
        filter (Tensor): a rank `n+2` dimensional array of kernel weights.
        strides (int, sequence of int, optional): a (sequence) of `n` integers,
            representing the inter-window strides. If a scalar is given, it is
            used `n` times. Defaults to `1`.
        padding (sequence of couple, `'SAME'`, `'VALID'`, optional): a sequence of
            `n` `(low, high)` integer pairs that give the padding to apply
            before and after each spatial dimension. For  `'VALID'`, those are
            `0`. For `'SAME'`, they are the `input length - filter length + 1`
            for each dim. Defaults to `'Valid'`.
        input_format (`None` or str, optional): a string of same length as the
            number of dimensions in `input` which specify their role
            (see below). Defaults to `'NCW'` for 1d conv, `'NCHW'` for 2d conv,
             and `'NDCHW'` for 3d conv.
        input_dilation (`None`, int or sequence of int, optional): giving the
            dilation factor to apply in each spatial dimension of `input`.
            Inumpy.t dilation is also known as transposed convolution as it allows
            to increase the output spatial dimension by inserting in the input
            any number of `0`s between each spatial value.
        filter_dilation (`None`, int or sequence of int): giving the dilation
            factor to apply in each spatial dimension of `filter`. Filter
            dilation is also known as atrous convolution as it corresponds to
            inserting any number of `0`s in between the filter values, similar
            to performing the non-dilated filter convolution with a subsample
            version of the input across the spatial dimensions.

    Returns:
        Tensor: An array containing the convolution result.

    Format of `input`, `filter` and `output`:
    For example, to indicate dimension numbers consistent with the `conv` function
    with two spatial dimensions, one could use `('NCHW', 'OIHW', 'NCHW')`. As
    another example, to indicate dimension numbers consistent with the TensorFlow
    Conv2D operation, one could use `('NHWC', 'HWIO', 'NHWC')`. When using the
    latter form of convolution dimension specification, window strides are
    associated with spatial dimension character labels according to the order in
    which the labels appear in the `rhs_spec` string, so that `window_strides[0]`
    is matched with the dimension corresponding to the first character
    appearing in rhs_spec that is not `'I'` or `'O'`.
    :param filter_format:
    :param output_format:
    """
    # setting up the strides
    if numpy.isscalar(strides):
        strides = (strides,) * (input.ndim - 2)
    elif len(strides) != (input.ndim - 2):
        msg = "given strides: {} should match the number".format(
            strides
        ) + "of spatial dim. in input: {}".format(input.ndim - 2)
        raise ValueError(msg)

    # setting up the padding
    if type(padding) != str:
        strides = (strides,) * (input.ndim - 2)
        if len(padding) != (input.ndim - 2):
            msg = "given padding: {} should match the ".format(
                padding
            ) + "number of spatial dim. in input: {}".format(input.ndim - 2)
            raise ValueError(msg)

    # setting up the filter_format
    if filter_format is None:
        if filter.ndim == 3:
            filter_format = "OIW"
        elif filter.ndim == 4:
            filter_format = "OIHW"
        elif filter.ndim == 5:
            filter_format = "OIDHW"
        else:
            msg = "filter_format should be given for >5 dimensions."
            raise ValueError(msg)
    elif len(filter_format) != filter.ndim:
        msg = "given filter_format: {} should".format(
            len(filter_format)
        ) + "match the number of dimension in filter: {}".format(filter.ndim)
        raise ValueError(msg)

    # setting up the input format
    if input_format is None:
        if len(filter.shape) == 3:
            input_format = "NCW"
        elif len(filter.shape) == 4:
            input_format = "NCHW"
        elif len(filter.shape) == 5:
            input_format = "NCDHW"
        else:
            msg = "input_format should be given for >5 dimensions."
            raise ValueError(msg)
    elif len(input_format) != input.ndim:
        msg = "given input_format: {} should".format(
            len(input_format)
        ) + "match the number of dimension in input: {}".format(input.ndim)
        raise ValueError(msg)

    # setting up the output format
    if output_format is None:
        if len(filter.shape) == 3:
            output_format = "NCW"
        elif len(filter.shape) == 4:
            output_format = "NCHW"
        elif len(filter.shape) == 5:
            output_format = "NCDHW"
        else:
            msg = "output_format should be given for >5 dimensions."
            raise ValueError(msg)
    elif len(output_format) != input.ndim:
        msg = "given output_format: {} should".format(
            len(output_format)
        ) + "match the number of dimension in output: {}".format(input.ndim)
        raise ValueError(msg)

    # setting up dilations
    if numpy.isscalar(input_dilation):
        input_dilation = (input_dilation,) * 2
    if numpy.isscalar(filter_dilation):
        filter_dilation = (filter_dilation,) * 2

    specs = (input_format, filter_format, output_format)
    return jla.conv_general_dilated(
        lhs=input,
        rhs=filter,
        window_strides=strides,
        padding=padding,
        lhs_dilation=input_dilation,
        rhs_dilation=filter_dilation,
        dimension_numbers=specs,
        precision=None,
    )


@jax_wrap
def convNd_transpose(
    input,
    filter,
    strides=1,
    padding="VALID",
    input_format=None,
    filter_format=None,
    output_format=None,
    input_dilation=None,
    filter_dilation=None,
    transpose_kernel=False,
):
    """General n-dimensional convolution operator, with optional dilation.

    Wraps Jax's conv_general_dilated functin, and thus also the XLA's `Conv
    <https://www.tensorflow.org/xla/operation_semantics#conv_convolution>`_
    operator.

    Args:
        input (Tensor): a rank `n+2` dimensional input array.
        filter (Tensor): a rank `n+2` dimensional array of kernel weights.
        strides (int, sequence of int, optional): a (sequence) of `n` integers,
            representing the inter-window strides. If a scalar is given, it is
            used `n` times. Defaults to `1`.
        padding (sequence of couple, `'SAME'`, `'VALID'`, optional): a sequence of
            `n` `(low, high)` integer pairs that give the padding to apply
            before and after each spatial dimension. For  `'VALID'`, those are
            `0`. For `'SAME'`, they are the `input length - filter length + 1`
            for each dim. Defaults to `'Valid'`.
        input_format (`None` or str, optional): a string of same length as the
            number of dimensions in `input` which specify their role
            (see below). Defaults to `'NCW'` for 1d conv, `'NCHW'` for 2d conv,
             and `'NDCHW'` for 3d conv.
        input_dilation (`None`, int or sequence of int, optional): giving the
            dilation factor to apply in each spatial dimension of `input`.
            Inumpy.t dilation is also known as transposed convolution as it allows
            to increase the output spatial dimension by inserting in the input
            any number of `0`s between each spatial value.
        filter_dilation (`None`, int or sequence of int): giving the dilation
            factor to apply in each spatial dimension of `filter`. Filter
            dilation is also known as atrous convolution as it corresponds to
            inserting any number of `0`s in between the filter values, similar
            to performing the non-dilated filter convolution with a subsample
            version of the input across the spatial dimensions.

    Returns:
        Tensor: An array containing the convolution result.

    Format of `input`, `filter` and `output`:
    For example, to indicate dimension numbers consistent with the `conv` function
    with two spatial dimensions, one could use `('NCHW', 'OIHW', 'NCHW')`. As
    another example, to indicate dimension numbers consistent with the TensorFlow
    Conv2D operation, one could use `('NHWC', 'HWIO', 'NHWC')`. When using the
    latter form of convolution dimension specification, window strides are
    associated with spatial dimension character labels according to the order in
    which the labels appear in the `rhs_spec` string, so that `window_strides[0]`
    is matched with the dimension corresponding to the first character
    appearing in rhs_spec that is not `'I'` or `'O'`.
    :param filter_format:
    :param output_format:
    :param transpose_kernel:
    """

    # setting up the strides
    if numpy.isscalar(strides):
        strides = (strides,) * (input.ndim - 2)
    elif len(strides) != (input.ndim - 2):
        msg = "given strides: {} should match the number".format(
            strides
        ) + "of spatial dim. in input: {}".format(input.ndim - 2)
        raise ValueError(msg)

    # setting up the padding
    if type(padding) != str:
        strides = (strides,) * (input.ndim - 2)
        if len(padding) != (input.ndim - 2):
            msg = "given padding: {} should match the ".format(
                padding
            ) + "number of spatial dim. in input: {}".format(input.ndim - 2)
            raise ValueError(msg)

    # setting up the filter_format
    if filter_format is None:
        if filter.ndim == 3:
            filter_format = "OIW"
        elif filter.ndim == 4:
            filter_format = "OIHW"
        elif filter.ndim == 5:
            filter_format = "OIDHW"
        else:
            msg = "filter_format should be given for >5 dimensions."
            raise ValueError(msg)
    elif len(filter_format) != filter.ndim:
        msg = "given filter_format: {} should".format(
            len(filter_format)
        ) + "match the number of dimension in filter: {}".format(filter.ndim)
        raise ValueError(msg)

    # setting up the input format
    if input_format is None:
        if len(filter.shape) == 3:
            input_format = "NCW"
        elif len(filter.shape) == 4:
            input_format = "NCHW"
        elif len(filter.shape) == 5:
            input_format = "NCDHW"
        else:
            msg = "input_format should be given for >5 dimensions."
            raise ValueError(msg)
    elif len(input_format) != input.ndim:
        msg = "given input_format: {} should".format(
            len(input_format)
        ) + "match the number of dimension in input: {}".format(input.ndim)
        raise ValueError(msg)

    # setting up the output format
    if output_format is None:
        if len(filter.shape) == 3:
            output_format = "NCW"
        elif len(filter.shape) == 4:
            output_format = "NCHW"
        elif len(filter.shape) == 5:
            output_format = "NCDHW"
        else:
            msg = "output_format should be given for >5 dimensions."
            raise ValueError(msg)
    elif len(output_format) != input.ndim:
        msg = "given output_format: {} should".format(
            len(output_format)
        ) + "match the number of dimension in output: {}".format(input.ndim)
        raise ValueError(msg)

    # setting up dilations
    if numpy.isscalar(input_dilation):
        input_dilation = (input_dilation,) * 2
    if numpy.isscalar(filter_dilation):
        filter_dilation = (filter_dilation,) * 2

    specs = (input_format, filter_format, output_format)
    return jla.conv_transpose(
        lhs=input,
        rhs=filter,
        strides=strides,
        padding=padding,
        rhs_dilation=filter_dilation,
        dimension_numbers=specs,
        precision=None,
        transpose_kernel=transpose_kernel,
    )


@jax_wrap
def poolNd(
    input,
    window_shape,
    reducer="MAX",
    strides=None,
    padding="VALID",
    init_val=None,
    rescalor=None,
):

    # set up the reducer
    if reducer == "MAX":
        reducer = jla.max
        rescalor = numpy.float32(1.0)
        init_val = -numpy.inf
    elif reducer == "SUM" or reducer == "AVG":
        reducer = jla.add
        if reducer == "AVG":
            rescalor = numpy.float32(1.0 / numpy.prod(window_shape))
        else:
            rescalor = numpy.float32(1.0)
        if init_val is None:
            init_val = 0.0

    # set up the window_shape
    if numpy.isscalar(window_shape):
        window_shape = (window_shape,) * input.ndim
    elif len(window_shape) != input.ndim:
        msg = "Given window_shape {} not the same length ".format(
            strides
        ) + "as input shape {}".format(input.ndim)
        raise ValueError(msg)

    # set up the strides
    if strides is None:
        strides = window_shape
    elif numpy.isscalar(strides):
        strides = (strides,) * len(window_shape)
    elif len(strides) != len(window_shape):
        msg = "Given strides {} not the same length ".format(
            strides
        ) + "as window_shape {}".format(window_shape)
        raise ValueError(msg)

    out = jla.reduce_window(
        operand=input * rescalor,
        init_value=init_val,
        computation=reducer,
        window_dimensions=window_shape,
        window_strides=strides,
        padding=padding,
    )
    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = "Randall Balestriero"

from .. import tensor
import numpy as np


_LOG_2PI = np.log(2 * np.pi)
_LOG_2 = np.log(2)
_LOG_PI = np.log(np.pi)


def psd(M):
    s, u = tensor.linalg.eigh(M, "L")
    # to do this function
    # s_pinv = np.array([0 if abs(x) <= eps else 1/x for x in v],
    #                    dtype=float)
    U = u * tensor.sqrt(1 / s)
    return U, tensor.log(s).sum()


class multivariate_normal:
    def logpdf(x, mean, cov):
        """
        Log of the multivariate normal probability density function.

        Parameters
        ----------
        x : array_like

        Returns
        -------
        pdf : ndarray
            Log of the probability density function evaluated at `x`
            :param mean:
            :param cov:

        """

        # we get the precision matrix such that
        # precision = dot(prec_U, prec_U.T)
        prec_U, log_det = psd(cov)
        centered_x = x - mean
        log_exp = (tensor.dot(centered_x, prec_U) ** 2).sum(axis=-1)
        return -0.5 * (x.shape[-1] * _LOG_2PI + log_det + log_exp)

        return _squeeze_output(out)

    def pdf(x, mean, cov):
        """
        Multivariate normal probability density function.
        Parameters
        ----------
        x : array_like
            Quantiles, with the last axis of `x` denoting the components.
        Returns
        -------
        pdf : ndarray
            Probability density function evaluated at `x`
            :param mean:
            :param cov:
        """
        return tensor.exp(multivariate_normal.logpdf(x, mean, cov))



from .base import Tensor, _add_method
import inspect
import sys

import jax.lax as jla
import jax.numpy as jnp

from .base import jax_wrap

module = sys.modules[__name__]

_JNP_NAMES = [c[0] for c in inspect.getmembers(jnp, inspect.isfunction)]
_TO_SKIP = [
    "<lambda>",
    "blackman",
    "bartlett",
    "hamming",
    "hanning",
    "kaiser",
    "add_docstring",
    "add_newdoc",
    "alen",
    "apply_along_axis",
    "apply_over_axes",
    "array2string",
    "array_equiv",
    "array_repr",
    "array_split",
    "array_str",
    "asanyarray",
    "asarray_chkfinite",
    "ascontiguousarray",
    "asfarray",
    "asfortranarray",
    "copy",
    "copyto",
    "custom_tra,nsforms",
    "delete",
    "deprecate",
    "device_put",
    "digitize",
    "disp",
    "function",
    "func",
    "find_common_type",
    "format_float_positional",
    "format_float_scientific",
    "frexp",
    "frombuffer",
    "fromfile",
    "fromfunction",
    "fromiter",
    "frompyfunc",
    "fromregex",
    "fromstring",
    "fv",
    "genfromtxt",
    "get_array_wrap",
    "get_include",
    "get_module_functions",
    "get_printoptions",
    "getbufsize",
    "geterr",
    "geterrcall",
    "geterrobj",
    "gradient",
    "int_asbuffer",
    "is_busday",
    "isrealobj",
    "issctype",
    "issubclass_",
    "issubdtype",
    "issubsctype",
    "iterable",
    "jit",
    "load",
    "loads",
    "loadtxt",
    "lookfor",
    "mafromtxt",
    "maximum_sctype",
    "may_share_memory",
    "mintypecode",
    "ndfromtxt",
    "nested_iters",
    "nextafter",
    "nper",
    "npv",
    "obj2sctype",
    "packbits",
    "printoptions",
    "recfromcsv",
    "recfromtxt",
    "removechars",
    "result_type",
    "safe_eval",
    "save",
    "savetxt",
    "savez",
    "savez_compressed",
    "sctype2char",
    "searchsorted",
    "select",
    "shape",
    "shares_memory",
    "show",
    "size",
    "source",
    "spacing",
    "strtobool",
    "trapz",
    "typename",
    "union1d",
    "unique",
    "unpackbits",
    "update_numpydoc",
    "who",
]

for name in _JNP_NAMES:
    if name in _TO_SKIP:
        continue
    module.__dict__.update({name: jax_wrap(jnp.__dict__[name])})

for name in [
    "issubsctype",
    "issubdtype",
    "nextafter",
    "result_type",
    "select",
    "isneginf",
    "isposinf",
    "logaddexp",
    "logaddexp2",
]:
    module.__dict__.update({name: jax_wrap(jnp.__dict__[name])})


cast = jax_wrap(jla.convert_element_type)
complex = jax_wrap(jla.complex)
range = module.__dict__["arange"]


def flatten(input):
    return module.__dict__["reshape"](input, (-1,))


def flatten2d(input):
    assert input.ndim > 1
    if input.ndim == 2:
        return input
    return reshape(input, (input.shape[0], -1))


################
getitem = jax_wrap(jnp.lax_numpy._rewriting_take)


_add_method(Tensor)(getitem, "__getitem__")

# overloading the basic arithmetic operators
_add_method(Tensor)(module.__dict__["add"], "__add__")
_add_method(Tensor)(module.__dict__["add"], "__radd__")
_add_method(Tensor)(module.__dict__["multiply"], "__mul__")
_add_method(Tensor)(module.__dict__["multiply"], "__rmul__")
_add_method(Tensor)(module.__dict__["true_divide"], "__truediv__")
_add_method(Tensor)(lambda a, b: module.__dict__["true_divide"](b, a), "__rtruediv__")
_add_method(Tensor)(module.__dict__["floor_divide"], "__floordiv__")
_add_method(Tensor)(lambda a, b: module.__dict__["floor_divide"](b, a), "__rfloordiv__")

_add_method(Tensor)(module.__dict__["subtract"], "__sub__")
_add_method(Tensor)(lambda a, b: module.__dict__["subtract"](b, a), "__rsub__")
_add_method(Tensor)(module.__dict__["power"], "__pow__")
_add_method(Tensor)(lambda a: 0 - a, "__neg__")

# overloading comparison operators
_add_method(Tensor)(module.__dict__["less"], "__lt__")
_add_method(Tensor)(module.__dict__["greater_equal"], "__rlt__")
_add_method(Tensor)(module.__dict__["greater"], "__gt__")
_add_method(Tensor)(module.__dict__["less_equal"], "__rgt__")
_add_method(Tensor)(module.__dict__["greater_equal"], "__ge__")
_add_method(Tensor)(module.__dict__["less"], "__rge__")
_add_method(Tensor)(module.__dict__["less_equal"], "__le__")
_add_method(Tensor)(module.__dict__["greater"], "__rle__")

# additional operators
_add_method(Tensor)(module.__dict__["sum"], "sum")
_add_method(Tensor)(module.__dict__["prod"], "prod")
_add_method(Tensor)(module.__dict__["mean"], "mean")
_add_method(Tensor)(module.__dict__["max"], "max")
_add_method(Tensor)(module.__dict__["min"], "min")
_add_method(Tensor)(module.__dict__["std"], "std")
_add_method(Tensor)(module.__dict__["var"], "var")
_add_method(Tensor)(module.__dict__["argmax"], "argmax")
_add_method(Tensor)(module.__dict__["argmin"], "argmin")

# additional operators

_add_method(Tensor)(module.__dict__["real"], "real")
_add_method(Tensor)(module.__dict__["imag"], "imag")
_add_method(Tensor)(module.__dict__["conjugate"], "conj")
_add_method(Tensor)(module.__dict__["conjugate"], "conjugate")
_add_method(Tensor)(module.__dict__["cast"], "cast")
_add_method(Tensor)(module.__dict__["cast"], "astype")
_add_method(Tensor)(module.__dict__["squeeze"], "squeeze")
_add_method(Tensor)(module.__dict__["flatten"], "flatten")
_add_method(Tensor)(module.__dict__["reshape"], "reshape")
_add_method(Tensor)(module.__dict__["dot"], "dot")
_add_method(Tensor)(module.__dict__["repeat"], "repeat")
_add_method(Tensor)(module.__dict__["expand_dims"], "expand_dims")
_add_method(Tensor)(module.__dict__["matmul"], "matmul")
_add_method(Tensor)(module.__dict__["round"], "round")
_add_method(Tensor)(module.__dict__["transpose"], "transpose")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = "Randall Balestriero"
import os
import pickle, gzip
import urllib.request
import numpy as np
import time
import h5py
from tqdm import tqdm


class birdvox_70k:
    """a dataset for avian flight call detection in half-second clips
    
    Version 1.0, April 2018.
    
    Created By
    
    Vincent Lostanlen (1, 2, 3), Justin Salamon (2, 3), Andrew Farnsworth (1),
    Steve Kelling (1), and Juan Pablo Bello (2, 3).
    
    (1): Cornell Lab of Ornithology (CLO)
    (2): Center for Urban Science and Progress, New York University
    (3): Music and Audio Research Lab, New York University
    
    https://wp.nyu.edu/birdvox
    
    Description
    
    The BirdVox-70k dataset contains 70k half-second clips from 6 audio
    recordings in the BirdVox-full-night dataset, each about ten hours in
    duration. These recordings come from ROBIN autonomous recording units,
    placed near Ithaca, NY, USA during the fall 2015. They were captured on the
    night of September 23rd, 2015, by six different sensors, originally
    numbered 1, 2, 3, 5, 7, and 10.
    
    Andrew Farnsworth used the Raven software to pinpoint every avian flight
    call in time and frequency. He found 35402 flight calls in total.
    He estimates that about 25 different species of passerines (thrushes,
    warblers, and sparrows) are present in this recording. Species are not
    labeled in BirdVox-70k, but it is possible to tell apart thrushes from
    warblers and sparrows by looking at the center frequencies of their calls.
    The annotation process took 102 hours.
    
    The dataset can be used, among other things, for the research,development
    and testing of bioacoustic classification models, including the
    reproduction of the results reported in [1].
    
    For details on the hardware of ROBIN recording units, we refer the reader
    to [2].
    
    [1] V. Lostanlen, J. Salamon, A. Farnsworth, S. Kelling, J. Bello. BirdVox-full-night: a dataset and benchmark for avian flight call detection. Proc. IEEE ICASSP, 2018.
    
    [2] J. Salamon, J. P. Bello, A. Farnsworth, M. Robbins, S. Keen, H. Klinck, and S. Kelling. Towards the Automatic Classification of Avian Flight Calls for Bioacoustic Monitoring. PLoS One, 2016.
    
    @inproceedings{lostanlen2018icassp,
    title = {BirdVox-full-night: a dataset and benchmark for avian flight call detection},
    author = {Lostanlen, Vincent and Salamon, Justin and Farnsworth, Andrew and Kelling, Steve and Bello, Juan Pablo},
    booktitle = {Proc. IEEE ICASSP},
    year = {2018},
    published = {IEEE},
    venue = {Calgary, Canada},
    month = {April},
    }
                   
    """

    @staticmethod
    def download(path):
        """
        Download the Birdvox dataset and store the result into the given
        path

        Parameters
        ----------

            path: str
                the path where the downloaded files will be stored. If the
                directory does not exist, it is created.
        """

        # Check if directory exists
        if not os.path.isdir(path + "birdvox_70k"):
            print("Creating birdvox_70k Directory")
            os.mkdir(path + "birdvox_70k")
        base = "https://zenodo.org/record/1226427/files/"
        basefile = "BirdVox-70k_unit{}.hdf5"
        names = ["01", "02", "03", "05", "07", "10"]
        # Check if file exists
        for name in names:
            filename = basefile.format(name)
            if not os.path.exists(path + "birdvox_70k/" + filename):
                url = base + filename + "?download=1"
                urllib.request.urlretrieve(url, path + "birdvox_70k/" + filename)

    @staticmethod
    def load(path=None):
        """
        Parameters
        ----------
            path: str (optional)
                default ($DATASET_PATH), the path to look for the data and
                where the data will be downloaded if not present

        Returns
        -------

            wavs: array(70804, 12000)
                the waveforms in the time amplitude domain

            labels: array(70804,)
                binary values representing the presence or not of an avian

            recording: array(70804,)
                the file number from which the sample has been extracted

        """

        if path is None:
            path = os.environ["DATASET_PATH"]

        birdvox_70k.download(path)

        t0 = time.time()

        # Loading the file
        path += "birdvox_70k/"
        names = ["01", "02", "03", "05", "07", "10"]
        basefile = "BirdVox-70k_unit{}.hdf5"
        wavs = list()
        label = list()
        recording = list()
        for name in names:
            f = h5py.File(path + basefile.format(name), "r")
            for filename in tqdm(
                f["waveforms"].keys(), ascii=True, desc="recording {}".format(name)
            ):
                wavs.append(f["waveforms"][filename][...])
                label.append(int(filename[-1]))
                recording.append(int(name))

        wavs = np.array(wavs).astype("float32")
        label = np.array(label).astype("int32")
        recording = np.array(recording).astype("int32")

        print("Dataset birdvox_70k loaded in {0:.2f}s.".format(time.time() - t0))

        return wavs, label, recording



import inspect
import sys

import jax
import jax.lax as jla
import jax.numpy as jnp
import numpy

from . import ops_numpy as T
from .base import jax_wrap
from ..nn.ops_nn import relu

module = sys.modules[__name__]

index = jax.ops.index


def hat_1D(x, t_left, t_center, t_right):
    """hat basis function in 1-D

    Hat function, continuous piecewise linear

    Parameters
    ----------

    x: array-like
        the sampled input space

    t_left: scalar
        the position of the left knot

    t_center: scalar
        the position of the center knot

    t_right: scalar
        the position of the right knot

    Returns
    -------

    output : array
        same shape as x with applied hat function
    """
    eps = 1e-6
    slope_left = 1 / (t_center - t_left)
    slope_right = 1 / (t_right - t_center)
    output = (
        (relu(x - t_left)) * slope_left
        - relu(x - t_center) * (slope_left + slope_right)
        + relu(x - t_right) * slope_right
    )
    return output


def _extract_signal_patches(signal, window_length, hop=1, data_format="NCW"):
    assert not hasattr(window_length, "__len__")
    assert signal.ndim == 3
    if data_format == "NCW":
        N = (signal.shape[2] - window_length) // hop + 1
        indices = jnp.arange(window_length) + jnp.expand_dims(jnp.arange(N) * hop, 1)
        indices = jnp.reshape(indices, [1, 1, N * window_length])
        patches = jnp.take_along_axis(signal, indices, 2)
        return jnp.reshape(patches, signal.shape[:2] + (N, window_length))
    else:
        error


extract_signal_patches = jax_wrap(_extract_signal_patches, module)


def _extract_image_patches(
    image, window_shape, hop=1, data_format="NCHW", mode="valid"
):
    if mode == "same":
        p1 = window_shape[0] - 1
        p2 = window_shape[1] - 1
        image = jnp.pad(
            image, [(0, 0), (0, 0), (p1 // 2, p1 - p1 // 2), (p2 // 2, p2 - p2 // 2)]
        )
    if not hasattr(hop, "__len__"):
        hop = (hop, hop)
    if data_format == "NCHW":

        # compute the number of windows in both dimensions
        N = (
            (image.shape[2] - window_shape[0]) // hop[0] + 1,
            (image.shape[3] - window_shape[1]) // hop[1] + 1,
        )

        # compute the base indices of a 2d patch
        patch = jnp.arange(numpy.prod(window_shape)).reshape(window_shape)
        offset = jnp.expand_dims(jnp.arange(window_shape[0]), 1)
        patch_indices = patch + offset * (image.shape[3] - window_shape[1])

        # create all the shifted versions of it
        ver_shifts = jnp.reshape(
            jnp.arange(N[0]) * hop[0] * image.shape[3], (-1, 1, 1, 1)
        )
        hor_shifts = jnp.reshape(jnp.arange(N[1]) * hop[1], (-1, 1, 1))
        all_cols = patch_indices + jnp.reshape(jnp.arange(N[1]) * hop[1], (-1, 1, 1))
        indices = patch_indices + ver_shifts + hor_shifts

        # now extract shape (1, 1, H'W'a'b')
        flat_indices = jnp.reshape(indices, [1, 1, -1])
        # shape is now (N, C, W*H)
        flat_image = jnp.reshape(image, (image.shape[0], image.shape[1], -1))
        # shape is now (N, C)
        patches = jnp.take_along_axis(flat_image, flat_indices, 2)
        return jnp.reshape(patches, image.shape[:2] + N + tuple(window_shape))
    else:
        error


extract_image_patches = jax_wrap(_extract_image_patches)


def _add_n(args):
    start = args[0]
    for arg in args:
        start = jnp.add(start, arg)
    return start


add_n = jax_wrap(_add_n)


def one_hot(i, N, dtype="float32"):
    """Create a one-hot encoding of x of size k."""
    if hasattr(i, "shape"):
        return (x[:, None] == arange(k)).astype(dtype)
    else:
        z = T.zeros(N, dtype)
        return index_add(z, i, 1)


for name in ["index_update", "index_min", "index_add", "index_max"]:
    module.__dict__.update({name: jax_wrap(jax.ops.__dict__[name])})

stop_gradient = jax_wrap(jla.stop_gradient)
dynamic_slice_in_dim = jax_wrap(jla.dynamic_slice_in_dim)
dynamic_slice = jax_wrap(jla.dynamic_slice)
index = jax.ops.index


module = sys.modules[__name__]

_NAMES = [c[0] for c in inspect.getmembers(jax.scipy.special, inspect.isfunction)]


for name in _NAMES:
    if name[0] == "_":
        continue
    module.__dict__.update({name: jax_wrap(jax.scipy.special.__dict__[name])})



# wavfile.py (Enhanced)
# Date: 20190213_2328 Joseph Ernest
#
# URL: https://gist.github.com/josephernest/3f22c5ed5dabf1815f16efa8fa53d476
# Source: scipy/io/wavfile.py
#
# Added:
# * read: also returns bitrate, cue markers + cue marker labels (sorted), loops, pitch
#         See https://web.archive.org/web/20141226210234/http://www.sonicspot.com/guide/wavefiles.html#labl
# * read: 24 bit & 32 bit IEEE files support (inspired from wavio_weckesser.py from Warren Weckesser)
# * read: added normalized (default False) that returns everything as float in [-1, 1]
# * read: added forcestereo that returns a 2-dimensional array even if input is mono
#
# * write: can write cue markers, cue marker labels, loops, pitch
# * write: 24 bit support
# * write: can write from a float normalized in [-1, 1]
# * write: 20180430_2335: bug fixed when size of data chunk is odd (previously, metadata could become unreadable because of this)
#
# * removed RIFX support (big-endian) (never seen one in 10+ years of audio production/audio programming), only RIFF (little-endian) are supported
# * removed read(..., mmap)
#
#
# Test:
# ..\wav\____wavfile_demo.py


"""
Module to read / write wav files using numpy arrays

Functions
---------
`read`: Return the sample rate (in samples/sec) and data from a WAV file.

`write`: Write a numpy array as a WAV file.

"""
from __future__ import division, print_function, absolute_import

import numpy
import struct
import warnings
import collections
from operator import itemgetter


class WavFileWarning(UserWarning):
    pass


_ieee = False

# assumes file pointer is immediately
#  after the 'fmt ' id
def _read_fmt_chunk(fid):
    res = struct.unpack("<ihHIIHH", fid.read(20))
    size, comp, noc, rate, sbytes, ba, bits = res
    if comp != 1 or size > 16:
        if comp == 3:
            global _ieee
            _ieee = True
            # warnings.warn("IEEE format not supported", WavFileWarning)
        else:
            warnings.warn("Unfamiliar format bytes", WavFileWarning)
        if size > 16:
            fid.read(size - 16)
    return size, comp, noc, rate, sbytes, ba, bits


# assumes file pointer is immediately
#   after the 'data' id
def _read_data_chunk(fid, noc, bits, normalized=False):
    size = struct.unpack("<i", fid.read(4))[0]

    if bits == 8 or bits == 24:
        dtype = "u1"
        bytes = 1
    else:
        bytes = bits // 8
        dtype = "<i%d" % bytes

    if bits == 32 and _ieee:
        dtype = "float32"

    data = numpy.fromfile(fid, dtype=dtype, count=size // bytes)

    if bits == 24:
        a = numpy.empty((len(data) // 3, 4), dtype="u1")
        a[:, :3] = data.reshape((-1, 3))
        a[:, 3:] = (a[:, 3 - 1 : 3] >> 7) * 255
        data = a.view("<i4").reshape(a.shape[:-1])

    if noc > 1:
        data = data.reshape(-1, noc)

    if bool(
        size & 1
    ):  # if odd number of bytes, move 1 byte further (data chunk is word-aligned)
        fid.seek(1, 1)

    if normalized:
        if bits == 8 or bits == 16 or bits == 24:
            normfactor = 2 ** (bits - 1)
        data = numpy.float32(data) * 1.0 / normfactor

    return data


def _skip_unknown_chunk(fid):
    data = fid.read(4)
    size = struct.unpack("<i", data)[0]
    if bool(
        size & 1
    ):  # if odd number of bytes, move 1 byte further (data chunk is word-aligned)
        size += 1
    fid.seek(size, 1)


def _read_riff_chunk(fid):
    str1 = fid.read(4)
    if str1 != b"RIFF":
        raise ValueError("Not a WAV file.")
    fsize = struct.unpack("<I", fid.read(4))[0] + 8
    str2 = fid.read(4)
    if str2 != b"WAVE":
        raise ValueError("Not a WAV file.")
    return fsize


def read(
    file,
    readmarkers=False,
    readmarkerlabels=False,
    readmarkerslist=False,
    readloops=False,
    readpitch=False,
    normalized=False,
    forcestereo=False,
):
    """
    Return the sample rate (in samples/sec) and data from a WAV file

    Parameters
    ----------
    file : file
        Input wav file.

    Returns
    -------
    rate : int
        Sample rate of wav file
    data : numpy array
        Data read from wav file

    Notes
    -----

    * The file can be an open file or a filename.

    * The returned sample rate is a Python integer
    * The data is returned as a numpy array with a
      data-type determined from the file.
      :param readmarkers:
      :param readmarkerlabels:
      :param readmarkerslist:
      :param readloops:
      :param readpitch:
      :param normalized:
      :param forcestereo:

    """
    if hasattr(file, "read"):
        fid = file
    else:
        fid = open(file, "rb")

    fsize = _read_riff_chunk(fid)
    noc = 1
    bits = 8
    # _cue = []
    # _cuelabels = []
    _markersdict = collections.defaultdict(lambda: {"position": -1, "label": ""})
    loops = []
    pitch = 0.0
    while fid.tell() < fsize:
        # read the next chunk
        chunk_id = fid.read(4)
        if chunk_id == b"fmt ":
            size, comp, noc, rate, sbytes, ba, bits = _read_fmt_chunk(fid)
        elif chunk_id == b"data":
            data = _read_data_chunk(fid, noc, bits, normalized)
        elif chunk_id == b"cue ":
            str1 = fid.read(8)
            size, numcue = struct.unpack("<ii", str1)
            for c in range(numcue):
                str1 = fid.read(24)
                (
                    id,
                    position,
                    datachunkid,
                    chunkstart,
                    blockstart,
                    sampleoffset,
                ) = struct.unpack("<iiiiii", str1)
                # _cue.append(position)
                _markersdict[id][
                    "position"
                ] = position  # needed to match labels and markers

        elif chunk_id == b"LIST":
            str1 = fid.read(8)
            size, type = struct.unpack("<ii", str1)
        elif chunk_id in [
            b"ICRD",
            b"IENG",
            b"ISFT",
            b"ISTJ",
        ]:  # see http://www.pjb.com.au/midi/sfspec21.html#i5
            _skip_unknown_chunk(fid)
        elif chunk_id == b"labl":
            str1 = fid.read(8)
            size, id = struct.unpack("<ii", str1)
            size = size + (
                size % 2
            )  # the size should be even, see WAV specfication, e.g. 16=>16, 23=>24
            label = fid.read(size - 4).rstrip(
                "\x00"
            )  # remove the trailing null characters
            # _cuelabels.append(label)
            _markersdict[id]["label"] = label  # needed to match labels and markers

        elif chunk_id == b"smpl":
            str1 = fid.read(40)
            (
                size,
                manuf,
                prod,
                sampleperiod,
                midiunitynote,
                midipitchfraction,
                smptefmt,
                smpteoffs,
                numsampleloops,
                samplerdata,
            ) = struct.unpack("<iiiiiIiiii", str1)
            cents = midipitchfraction * 1.0 / (2 ** 32 - 1)
            pitch = 440.0 * 2 ** ((midiunitynote + cents - 69.0) / 12)
            for i in range(numsampleloops):
                str1 = fid.read(24)
                cuepointid, type, start, end, fraction, playcount = struct.unpack(
                    "<iiiiii", str1
                )
                loops.append([start, end])
        else:
            warnings.warn("Chunk " + chunk_id + " skipped", WavFileWarning)
            _skip_unknown_chunk(fid)
    fid.close()

    if data.ndim == 1 and forcestereo:
        data = numpy.column_stack((data, data))

    _markerslist = sorted(
        [_markersdict[l] for l in _markersdict], key=lambda k: k["position"]
    )  # sort by position
    _cue = [m["position"] for m in _markerslist]
    _cuelabels = [m["label"] for m in _markerslist]

    return (
        (rate, data, bits,)
        + ((_cue,) if readmarkers else ())
        + ((_cuelabels,) if readmarkerlabels else ())
        + ((_markerslist,) if readmarkerslist else ())
        + ((loops,) if readloops else ())
        + ((pitch,) if readpitch else ())
    )


def write(
    filename,
    rate,
    data,
    bitrate=None,
    markers=None,
    loops=None,
    pitch=None,
    normalized=False,
):
    """
    Write a numpy array as a WAV file

    Parameters
    ----------
    filename : file
        The name of the file to write (will be over-written).
    rate : int
        The sample rate (in samples/sec).
    data : ndarray
        A 1-D or 2-D numpy array of integer data-type.

    Notes
    -----
    * Writes a simple uncompressed WAV file.
    * The bits-per-sample will be determined by the data-type.
    * To write multiple-channels, use a 2-D array of shape
      (Nsamples, Nchannels).
      :param bitrate:
      :param markers:
      :param loops:
      :param pitch:
      :param normalized:

    """

    # normalization and 24-bit handling
    if (
        bitrate == 24
    ):  # special handling of 24 bit wav, because there is no numpy.int24...
        if normalized:
            data[data > 1.0] = 1.0
            data[data < -1.0] = -1.0
            a32 = numpy.asarray(data * (2 ** 23 - 1), dtype=numpy.int32)
        else:
            a32 = numpy.asarray(data, dtype=numpy.int32)
        if a32.ndim == 1:
            a32.shape = a32.shape + (1,)  # Convert to a 2D array with a single column.
        a8 = (
            a32.reshape(a32.shape + (1,)) >> numpy.array([0, 8, 16])
        ) & 255  # By shifting first 0 bits, then 8, then 16, the resulting output is 24 bit little-endian.
        data = a8.astype(numpy.uint8)
    else:
        if normalized:  # default to 32 bit int
            data[data > 1.0] = 1.0
            data[data < -1.0] = -1.0
            data = numpy.asarray(data * (2 ** 31 - 1), dtype=numpy.int32)

    fid = open(filename, "wb")
    fid.write(b"RIFF")
    fid.write(b"\x00\x00\x00\x00")
    fid.write(b"WAVE")

    # fmt chunk
    fid.write(b"fmt ")
    if data.ndim == 1:
        noc = 1
    else:
        noc = data.shape[1]
    bits = data.dtype.itemsize * 8 if bitrate != 24 else 24
    sbytes = rate * (bits // 8) * noc
    ba = noc * (bits // 8)
    fid.write(struct.pack("<ihHIIHH", 16, 1, noc, rate, sbytes, ba, bits))

    # cue chunk
    if markers:  # != None and != []
        if isinstance(
            markers[0], dict
        ):  # then we have [{'position': 100, 'label': 'marker1'}, ...]
            labels = [m["label"] for m in markers]
            markers = [m["position"] for m in markers]
        else:
            labels = ["" for m in markers]

        fid.write(b"cue ")
        size = 4 + len(markers) * 24
        fid.write(struct.pack("<ii", size, len(markers)))
        for i, c in enumerate(markers):
            s = struct.pack(
                "<iiiiii", i + 1, c, 1635017060, 0, 0, c
            )  # 1635017060 is struct.unpack('<i',b'data')
            fid.write(s)

        lbls = ""
        for i, lbl in enumerate(labels):
            lbls += b"labl"
            label = lbl + ("\x00" if len(lbl) % 2 == 1 else "\x00\x00")
            size = len(lbl) + 1 + 4  # because \x00
            lbls += struct.pack("<ii", size, i + 1)
            lbls += label

        fid.write(b"LIST")
        size = len(lbls) + 4
        fid.write(struct.pack("<i", size))
        fid.write(
            b"adtl"
        )  # https://web.archive.org/web/20141226210234/http://www.sonicspot.com/guide/wavefiles.html#list
        fid.write(lbls)

    # smpl chunk
    if loops or pitch:
        if not loops:
            loops = []
        if pitch:
            midiunitynote = 12 * numpy.log2(pitch * 1.0 / 440.0) + 69
            midipitchfraction = int(
                (midiunitynote - int(midiunitynote)) * (2 ** 32 - 1)
            )
            midiunitynote = int(midiunitynote)
            # print(midipitchfraction, midiunitynote)
        else:
            midiunitynote = 0
            midipitchfraction = 0
        fid.write(b"smpl")
        size = 36 + len(loops) * 24
        sampleperiod = int(1000000000.0 / rate)

        fid.write(
            struct.pack(
                "<iiiiiIiiii",
                size,
                0,
                0,
                sampleperiod,
                midiunitynote,
                midipitchfraction,
                0,
                0,
                len(loops),
                0,
            )
        )
        for i, loop in enumerate(loops):
            fid.write(struct.pack("<iiiiii", 0, 0, loop[0], loop[1], 0, 0))

    # data chunks
    fid.write(b"data")
    fid.write(struct.pack("<i", data.nbytes))
    import sys

    if data.dtype.byteorder == ">" or (
        data.dtype.byteorder == "=" and sys.byteorder == "big"
    ):
        data = data.byteswap()

    data.tofile(fid)

    if (
        data.nbytes % 2 == 1
    ):  # add an extra padding byte if data.nbytes is odd: https://web.archive.org/web/20141226210234/http://www.sonicspot.com/guide/wavefiles.html#data
        fid.write("\x00")

    # Determine file size and place it in correct
    #  position at start of the file.
    size = fid.tell()
    fid.seek(4)
    fid.write(struct.pack("<i", size - 8))
    fid.close()



import numpy as np


def constant(shape, value):
    return np.full(shape, value)


def uniform(shape, range=0.01, std=None, mean=0.0):
    """Sample initial weights from the uniform distribution.

    Parameters are sampled from U(a, b).

    Parameters
    ----------

    range: float or tuple
        When std is None then range determines a, b. If range is a float the
        weights are sampled from U(-range, range). If range is a tuple the
        weights are sampled from U(range[0], range[1]).

    std: float or None
        If std is a float then the weights are sampled from
        U(mean - np.sqrt(3) * std, mean + np.sqrt(3) * std).

    mean: float
        see std for description.
        :param shape:
    """
    if std is not None:
        a = mean - np.sqrt(3) * std
        b = mean + np.sqrt(3) * std
    elif hasattr(range, "__len__"):
        a, b = range  # range is a tuple
    else:
        a, b = -range, range  # range is a number
    return np.random.rand(*shape) * (b - a) + a


def normal(shape, mean=0.0, std=1.0):
    """Sample initial weights from the Gaussian distribution.

    Initial weight parameters are sampled from N(mean, std).

    Parameters
    ----------

    std: float
        Std of initial parameters.

    mean: float
        Mean of initial parameters.
        :param shape:

    """
    return np.random.randn(*shape) * std + mean


def orthogonal(shape, gain=1):
    flat_shape = (shape[0], np.prod(shape[1:]))
    a = np.random.normal(flat_shape)
    u, _, v = np.linalg.svd(a, full_matrices=False)
    # pick the one with the correct shape
    q = u if u.shape == flat_shape else v
    q = q.reshape(shape)
    return gain * q


def _compute_fans(shape, in_axis, out_axis):
    fan_in = np.prod([shape[i] for i in in_axis])
    fan_out = np.prod([shape[i] for i in out_axis])
    return fan_in, fan_out


def variance_scaling(
    mode, shape, gain, distribution=normal, in_axis=None, out_axis=None
):
    """Variance Scaling initialization.
    """

    if in_axis is None:
        if len(shape) == 2:
            in_axis = [0]
        else:
            in_axis = list(range(1, len(shape)))

    if out_axis is None:
        out_axis = [i for i in range(len(shape)) if i not in in_axis]

    if len(shape) < 2:
        raise RuntimeError("This initializer only works with shapes of length >= 2")

    fan_in, fan_out = _compute_fans(shape, in_axis, out_axis)
    if mode == "fan_in":
        den = fan_in
    elif mode == "fan_out":
        den = fan_out
    elif mode == "fan_avg":
        den = (fan_in + fan_out) / 2.0
    else:
        raise ValueError(
            "mode must be fan_in, fan_out or fan_avg, value passed was {mode}"
        )
    std = gain * np.sqrt(1.0 / den)
    return distribution(shape, std=std)


def glorot(shape, gain=1, distribution=normal, in_axis=None, out_axis=None):
    """Glorot weight initialization.

    This is also known as Xavier initialization [1]_.

    Parameters
    ----------

    initializer: lasagne.init.Initializer
        Initializer used to sample the weights, must accept `std` in its
        constructor to sample from a distribution with a given standard
        deviation.

    gain: float or 'relu'
        Scaling factor for the weights. Set this to ``1.0`` for linear and
        sigmoid units, to 'relu' or ``sqrt(2)`` for rectified linear units, and
        to ``sqrt(2/(1+alpha**2))`` for leaky rectified linear units with
        leakiness ``alpha``. Other transfer functions may need different
        factors.

    c01b: bool
        For a :class:`lasagne.layers.cuda_convnet.Conv2DCCLayer` constructed
        with ``dimshuffle=False``, `c01b` must be set to ``True`` to compute
        the correct fan-in and fan-out.

    References
    ----------

    .. [1] Xavier Glorot and Yoshua Bengio (2010):
           Understanding the difficulty of training deep feedforward neural
           networks. International conference on artificial intelligence and
           statistics.

    Notes
    -----

    For a :class:`DenseLayer <lasagne.layers.DenseLayer>`, if ``gain='relu'``
    and ``initializer=Uniform``, the weights are initialized as
    .. math::
       a &= \\sqrt{\\frac{12}{fan_{in}+fan_{out}}}\\\\
       W &\sim U[-a, a]
    If ``gain=1`` and ``initializer=Normal``, the weights are initialized as
    .. math::
       \\sigma &= \\sqrt{\\frac{2}{fan_{in}+fan_{out}}}\\\\
       W &\sim N(0, \\sigma)
       :param shape:
       :param distribution:
       :param in_axis:
       :param out_axis:
    """
    if len(shape) < 2:
        raise RuntimeError("This initializer only works with shapes of length >= 2")

    return variance_scaling(
        "fan_avg", shape, gain, distribution, in_axis=in_axis, out_axis=out_axis,
    )


def he(shape, gain=np.sqrt(2), distribution=normal, in_axis=None, out_axis=None):
    """He weight initialization.
    Weights are initialized with a standard deviation of
    :param shape:
    :param distribution:
    :param in_axis:
    :param out_axis:
    :math:`\\sigma = gain \\sqrt{\\frac{1}{fan_{in}}}` [1]_.

    Parameters
    ----------

    initializer : lasagne.init.Initializer
        Initializer used to sample the weights, must accept `std` in its
        constructor to sample from a distribution with a given standard
        deviation.

    gain : float or 'relu'
        Scaling factor for the weights. Set this to ``1.0`` for linear and
        sigmoid units, to 'relu' or ``sqrt(2)`` for rectified linear units, and
        to ``sqrt(2/(1+alpha**2))`` for leaky rectified linear units with
        leakiness ``alpha``. Other transfer functions may need different
        factors.

    c01b : bool
        For a :class:`lasagne.layers.cuda_convnet.Conv2DCCLayer` constructed
        with ``dimshuffle=False``, `c01b` must be set to ``True`` to compute
        the correct fan-in and fan-out.

    References
    ----------

    .. [1] Kaiming He et al. (2015):
           Delving deep into rectifiers: Surpassing human-level performance on
           imagenet classification. arXiv preprint arXiv:1502.01852.

    See Also
    ----------

    HeNormal  : Shortcut with Gaussian initializer.
    HeUniform : Shortcut with uniform initializer.
    """
    return variance_scaling(
        "fan_in", shape, gain, distribution, in_axis=in_axis, out_axis=out_axis
    )


def lecun(shape, gain=1.0, distribution=normal, in_axis=None, out_axis=None):
    """LeCun weight initialization.
    Weights are initialized with a standard deviation of
    :math:`\\sigma = gain \\sqrt{\\frac{1}{fan_{in}}}`.
   """
    return variance_scaling(
        "fan_in", shape, gain, distribution, in_axis=in_axis, out_axis=out_axis
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from symjax import tensor as T
from ..base import current_graph, Scope


def ExponentialMovingAverage(value, alpha):

    with Scope("ExponentialMovingAverage"):

        first_step = T.Variable(True, trainable=False, name="first_step", dtype="bool")

        var = T.Variable(
            T.zeros(value.shape), trainable=False, dtype="float32", name="EMA"
        )

        new_value = T.where(first_step, value, var * alpha + (1 - alpha) * value)

        current_graph().add({var: new_value, first_step: False})

    return new_value, var


def PiecewiseConstant(init, steps_and_values):

    with Scope("PiecewiseConstant"):

        all_steps = T.stack([0] + list(steps_and_values.keys()))
        all_values = T.stack([init] + list(steps_and_values.values()))

        step = T.Variable(T.zeros(1), trainable=False, name="step", dtype="float32",)

        value = all_values[(step < all_steps).argmin() - 1]

        current_graph().add({step: step + 1})

    return value
